""" Test the FilenamePlugin class"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
from DIRAC.Resources.Catalog.ConditionPlugins.FilenamePlugin import FilenamePlugin

__RCSID__ = "$Id $"


class TestfilenamePlugin(unittest.TestCase):
    """Test the FilenamePlugin class"""

    def setUp(self):
        self.lfns = ["/lhcb/lfn1", "/lhcb/anotherlfn", "/otherVo/name"]

    def test_01_endswith(self):
        """Testing endswith (method with argument"""

        fnp = FilenamePlugin("endswith('n')")

        self.assertTrue(not fnp.eval(lfn="/lhcb/lfn1"))
        self.assertTrue(fnp.eval(lfn="/lhcb/lfn"))

    def test_02_find(self):
        """Testing special case of find"""

        fnp = FilenamePlugin("find('lfn')")

        self.assertTrue(fnp.eval(lfn="/lhcb/lfn1"))
        self.assertTrue(not fnp.eval(lfn="/lhcb/l0f0n"))

    def test_03_isalnum(self):
        """Testing isalnum (method without argument"""

        fnp = FilenamePlugin("isalnum()")

        self.assertTrue(fnp.eval(lfn="lhcblfn1"))
        self.assertTrue(not fnp.eval(lfn="/lhcb/lf_n"))

    def test_04_nonExisting(self):
        """Testing non existing string method"""

        fnp = FilenamePlugin("nonexisting()")

        self.assertTrue(not fnp.eval(lfn="lhcblfn1"))
        self.assertTrue(not fnp.eval(lfn="/lhcb/lf_n"))


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestfilenamePlugin)

    unittest.TextTestRunner(verbosity=2).run(suite)



""" PilotsLoggingHandler is the implementation of the PilotsLogging service

    The following methods are available in the Service interface

    addPilotsLogging()
    getPilotsLogging
    deletePilotsLogging()

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.Core.DISET.RequestHandler import RequestHandler

from DIRAC.Resources.MessageQueue.MQCommunication import createConsumer


class PilotsLoggingHandler(RequestHandler):
    """Server side functions for Pilots Logging service"""

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """Initialization of Pilots Logging service"""
        cls.consumersSet = set()
        try:
            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.PilotsLoggingDB", "PilotsLoggingDB")
            if not result["OK"]:
                return result
            cls.pilotsLoggingDB = result["Value"]()

        except RuntimeError as excp:
            return S_ERROR("Can't connect to DB: %s" % excp)

        queue = cls.srv_getCSOption("PilotsLoggingQueue")
        # This is pretty awful hack. Somehow, for uknown reason, I cannot access CS with srv_getCSOption.
        # The only way is using full CS path, so I'm using it as a backup solution.
        if not queue:
            queue = gConfig.getValue(serviceInfoDict["serviceSectionPath"] + "/PilotsLoggingQueue")
        result = createConsumer(queue, callback=cls.consumingCallback)
        if result["OK"]:
            cls.consumersSet.add(result["Value"])
        else:
            return result
        return S_OK()

    @classmethod
    def consumingCallback(cls, headers, message):
        """
        Callback function for the MQ Consumer, called for every new message and inserting it into database.

        :param headers: Headers of MQ message (not used)
        :param message: Message represented as a dictionary
        """
        # verify received message format
        if set(message) == set(["pilotUUID", "timestamp", "source", "phase", "status", "messageContent"]):
            cls.pilotsLoggingDB.addPilotsLogging(
                message["pilotUUID"],
                message["timestamp"],
                message["source"],
                message["phase"],
                message["status"],
                message["messageContent"],
            )

    types_addPilotsLogging = [
        six.string_types,
        six.string_types,
        six.string_types,
        six.string_types,
        six.string_types,
        six.string_types,
    ]

    @classmethod
    def export_addPilotsLogging(cls, pilotUUID, timestamp, source, phase, status, messageContent):
        """
        Add new Pilots Logging entry

        :param pilotUUID: Pilot reference
        :param status: Pilot status
        :param minorStatus: Additional status information
        :param timeStamp: Date and time of status event
        :param source: Source of statu information
        """

        return cls.pilotsLoggingDB.addPilotsLogging(pilotUUID, timestamp, source, phase, status, messageContent)

    types_getPilotsLogging = [six.string_types]

    @classmethod
    def export_getPilotsLogging(cls, pilotUUID):
        """
        Get all Logging entries for Pilot

        :param pilotUUID: Pilot reference
        """

        return cls.pilotsLoggingDB.getPilotsLogging(pilotUUID)

    types_deletePilotsLogging = [six.string_types + (list,)]

    @classmethod
    def export_deletePilotsLogging(cls, pilotUUID):
        """
        Delete all Logging entries for Pilot

        :param pilotUUID: Pilot reference
        """

        return cls.pilotsLoggingDB.deletePilotsLogging(pilotUUID)



""" The Matcher service provides an interface for matching jobs to pilots

    It uses a Matcher and a Limiter object that encapsulates the matching logic.
    It connects to JobDB, TaskQueueDB, JobLoggingDB, and PilotAgentsDB.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six

from DIRAC import S_OK, S_ERROR

from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities.DEncode import ignoreEncodeWarning
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader

from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor

from DIRAC.WorkloadManagementSystem.Client.Matcher import Matcher, PilotVersionError
from DIRAC.WorkloadManagementSystem.Client.Limiter import Limiter


class MatcherHandlerMixin(object):
    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        try:
            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.JobDB", "JobDB")
            if not result["OK"]:
                return result
            cls.jobDB = result["Value"]()

            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.JobLoggingDB", "JobLoggingDB")
            if not result["OK"]:
                return result
            cls.jobLoggingDB = result["Value"]()

            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.TaskQueueDB", "TaskQueueDB")
            if not result["OK"]:
                return result
            cls.taskQueueDB = result["Value"]()

            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.PilotAgentsDB", "PilotAgentsDB")
            if not result["OK"]:
                return result
            cls.pilotAgentsDB = result["Value"]()

        except RuntimeError as excp:
            return S_ERROR("Can't connect to DB: %s" % excp)

        cls.limiter = Limiter(jobDB=cls.jobDB)

        gMonitor.registerActivity("matchTime", "Job matching time", "Matching", "secs", gMonitor.OP_MEAN, 300)
        gMonitor.registerActivity("matchesDone", "Job Match Request", "Matching", "matches", gMonitor.OP_RATE, 300)
        gMonitor.registerActivity("matchesOK", "Matched jobs", "Matching", "matches", gMonitor.OP_RATE, 300)
        gMonitor.registerActivity("numTQs", "Number of Task Queues", "Matching", "tqsk queues", gMonitor.OP_MEAN, 300)

        return S_OK()

    ##############################################################################
    types_requestJob = [six.string_types + (dict,)]

    def export_requestJob(self, resourceDescription):
        """Serve a job to the request of an agent which is the highest priority
        one matching the agent's site capacity
        """

        resourceDescription["Setup"] = self.serviceInfoDict["clientSetup"]
        credDict = self.getRemoteCredentials()
        pilotRef = resourceDescription.get("PilotReference", "Unknown")

        try:
            opsHelper = Operations(group=credDict["group"])
            matcher = Matcher(
                pilotAgentsDB=self.pilotAgentsDB,
                jobDB=self.jobDB,
                tqDB=self.taskQueueDB,
                jlDB=self.jobLoggingDB,
                opsHelper=opsHelper,
                pilotRef=pilotRef,
            )
            result = matcher.selectJob(resourceDescription, credDict)
        except RuntimeError as rte:
            self.log.error("Error requesting job for pilot", "[%s] %s" % (pilotRef, rte))
            return S_ERROR("Error requesting job")
        except PilotVersionError as pve:
            self.log.warn("Pilot version error for pilot", "[%s] %s" % (pilotRef, pve))
            return S_ERROR("Error requesting job")

        # result can be empty, meaning that no job matched
        if result:
            gMonitor.addMark("matchesDone")
            gMonitor.addMark("matchesOK")
            return S_OK(result)
        # FIXME: This is correctly interpreted by the JobAgent, but DErrno should be used instead
        return S_ERROR("No match found")

    ##############################################################################
    types_getActiveTaskQueues = []

    @classmethod
    @ignoreEncodeWarning
    def export_getActiveTaskQueues(cls):
        """Return all task queues"""
        return cls.taskQueueDB.retrieveTaskQueues()

    ##############################################################################
    types_getMatchingTaskQueues = [dict]

    @classmethod
    @ignoreEncodeWarning
    def export_getMatchingTaskQueues(cls, resourceDict):
        """Return all task queues that match the resourceDict"""
        if "Site" in resourceDict and isinstance(resourceDict["Site"], six.string_types):
            gridCE = resourceDict.get("GridCE")
            negativeCond = cls.limiter.getNegativeCondForSite(resourceDict["Site"], gridCE)
        else:
            negativeCond = cls.limiter.getNegativeCond()
        matcher = Matcher(pilotAgentsDB=cls.pilotAgentsDB, jobDB=cls.jobDB, tqDB=cls.taskQueueDB, jlDB=cls.jobLoggingDB)
        resourceDescriptionDict = matcher._processResourceDescription(resourceDict)
        return cls.taskQueueDB.getMatchingTaskQueues(resourceDescriptionDict, negativeCond=negativeCond)


class MatcherHandler(MatcherHandlerMixin, RequestHandler):
    pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


from base64 import b64encode, b64decode

from DIRAC.Core.Tornado.Client.TornadoClient import TornadoClient
from DIRAC.Core.Base.Client import Client


class CSJSONClient(TornadoClient):
    """
    The specific Tornado client for configuration system.
    To avoid JSON limitation the HTTPS handler encode data in base64
    before sending them, this class only decode the base64

    An exception is made with CommitNewData which ENCODE in base64
    """

    def getCompressedData(self):
        """
        Transmit request to service and get data in base64,
        it decode base64 before returning

        :returns: Configuration data, compressed
        :rtype: str
        """
        retVal = self.executeRPC("getCompressedData")
        if retVal["OK"]:
            retVal["Value"] = b64decode(retVal["Value"])
        return retVal

    def getCompressedDataIfNewer(self, sClientVersion):
        """
        Transmit request to service and get data in base64,
        it decode base64 before returning.

        :returns: Configuration data, if changed, compressed
        """
        retVal = self.executeRPC("getCompressedDataIfNewer", sClientVersion)
        if retVal["OK"] and "data" in retVal["Value"]:
            retVal["Value"]["data"] = b64decode(retVal["Value"]["data"])
        return retVal

    def commitNewData(self, sData):
        """
        Transmit request to service by encoding data in base64.

        :param sData: Data to commit, you may call this method with CSAPI and Modificator
        """
        return self.executeRPC("commitNewData", b64encode(sData).decode())


class ConfigurationClient(Client):
    """
    Specific client to speak with ConfigurationServer.

    This class must contain at least the JSON decoder dedicated to
    the Configuration Server.

    You can implement more logic or function to the client here if needed,
    RPCCall can be made inside this class with executeRPC method.
    """

    # The JSON decoder for Configuration Server
    httpsClient = CSJSONClient

    def __init__(self, **kwargs):
        # By default we use Configuration/Server as url, client do the resolution
        # In some case url has to be static (when a slave register to the master server for example)
        # It's why we can use 'url' as keyword arguments
        if "url" not in kwargs:
            kwargs["url"] = "Configuration/Server"
        super(ConfigurationClient, self).__init__(**kwargs)



""" SlackAction

  This action posts the status change to a Slack channel
  It uses Slack Webhooks.
  Also compatible with Mattermost, which is an open source alternative to Slack.
  To create a webhook URL refer the following :

  * Slack : https://api.slack.com/incoming-webhooks
  * Mattermost : https://docs.mattermost.com/developer/webhooks-incoming.html

  Add the webhook URL to dirac.cfg at Operations/[]/ResourceStatus/Config/Slack

  example:

    Operations/
      Defaults/
        ResourceStatus/
          Config/
            Slack = https://hooks.slack.com/services/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
            Mattermost = https://mattermost.web.cern.ch/hooks/axy94k3m1pg5xeyaw3qqb3x8bo

  Even if using Mattermost,the URL is still to be placed at
  Operations/[]/ResourceStatus/Config/Slack and not Operations/[]/ResourceStatus/Config/Mattermost

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import json
import requests
import six

from DIRAC import S_ERROR, S_OK
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ResourceStatusSystem.PolicySystem.Actions.BaseAction import BaseAction
from DIRAC.Core.Utilities.SiteSEMapping import getSitesForSE
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getCESiteMapping


class SlackAction(BaseAction):
    """
    Action that sends a brief Slack Message.
    """

    def __init__(self, name, decisionParams, enforcementResult, singlePolicyResults, clients=None, url=None):

        super(SlackAction, self).__init__(name, decisionParams, enforcementResult, singlePolicyResults, clients)
        if url is not None:
            self.url = url
        else:
            self.url = Operations().getValue("ResourceStatus/Config/Slack")

    def run(self):
        """
        Checks it has the parameters it needs and tries to send an sms to the users
        that apply.
        """

        if self.url is None:
            return S_ERROR("Slack URL not set")

        # Minor security checks
        element = self.decisionParams["element"]
        if element is None:
            return S_ERROR("element should not be None")

        name = self.decisionParams["name"]
        if name is None:
            return S_ERROR("name should not be None")

        statusType = self.decisionParams["statusType"]
        if statusType is None:
            return S_ERROR("statusType should not be None")

        previousStatus = self.decisionParams["status"]
        if previousStatus is None:
            return S_ERROR("status should not be None")

        status = self.enforcementResult["Status"]
        if status is None:
            return S_ERROR("status should not be None")

        reason = self.enforcementResult["Reason"]
        if reason is None:
            return S_ERROR("reason should not be None")

        if self.decisionParams["element"] == "Site":
            siteName = self.decisionParams["name"]
        else:
            elementType = self.decisionParams["elementType"]

            if elementType == "StorageElement":
                siteName = getSitesForSE(name)
            elif elementType == "ComputingElement":
                res = getCESiteMapping(name)
                if not res["OK"]:
                    self.log.error("Failure getting Site2CE mapping", res["Message"])
                    siteName = "ERROR"
                else:
                    siteName = res
            else:
                siteName = {"OK": True, "Value": "Unassigned"}

            if not siteName["OK"]:
                self.log.error("Resource %s does not exist at any site: %s" % (name, siteName["Message"]))
                siteName = "Unassigned Resources"
            elif not siteName["Value"]:
                siteName = "Unassigned Resources"
            else:
                siteName = (
                    siteName["Value"] if isinstance(siteName["Value"], six.string_types) else siteName["Value"][0]
                )

        message = "*{name}* _{statusType}_ --> _{status}_ \n{reason}".format(
            name=name, statusType=statusType, status=status, reason=reason
        )
        return self.sendSlackMessage(message)

    def sendSlackMessage(self, message):
        """
        Sends a slack message to self.url

        :param str message: text message to send
        """

        payload = {"text": message}
        response = requests.post(self.url, data=json.dumps(payload), headers={"Content-Type": "application/json"})
        response.raise_for_status()
        return S_OK()



"""
    Workflow Utility module contains a number of functions useful for various
    workflow operations
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import six
import re


def getSubstitute(param, skip_list=[]):
    """Get the variable name to which the given parameter is referring"""
    result = []
    resList = re.findall(r"@{([][\w,.:$()]+)}", str(param))
    if resList:
        for match in resList:
            if match not in skip_list:
                result.append(match)

    return result


def substitute(param, variable, value):
    """Substitute the variable reference with the value"""

    tmp_string = str(param).replace("@{" + variable + "}", value)
    if isinstance(param, six.string_types):
        return tmp_string
    return eval(tmp_string)


def resolveVariables(varDict):
    """Resolve variables defined in terms of others within the same dictionary"""
    max_tries = 10
    variables = list(varDict)
    ntry = 0
    while ntry < max_tries:
        substFlag = False
        for var, value in list(varDict.items()):
            if isinstance(value, six.string_types):
                substitute_vars = getSubstitute(value)
                for substitute_var in substitute_vars:
                    if substitute_var in variables:
                        varDict[var] = substitute(varDict[var], substitute_var, varDict[substitute_var])
                        substFlag = True
        if not substFlag:
            break
        ntry += 1
    else:
        print("Failed to resolve referencies in %d attempts" % max_tries)


def dataFromOption(parameter):

    result = []

    if parameter.type.lower() == "option":

        fields = parameter.value.split(",")

        for f in fields:
            if re.search(r"FILE\s*=", f):
                fname = re.search(r"FILE\s*=\s*'([][;\/\w.:\s@{}-]+)'", f).group(1)
                res = re.search(r"TYP\w*\s*=\s*'(\w+)'", f)
                if res:
                    ftype = res.group(1)
                else:
                    ftype = "Unknown"

                result.append((fname, ftype))

    return result


def expandDatafileOption(option):

    result = ""

    if not re.search(";;", option.value):
        return result

    files = dataFromOption(option)
    if len(files) == 1:
        fname, ftype = files[0]
        fnames = fname.split(";;")
        if len(fnames) > 1:

            template = option.value.strip().replace("=", "", 1)
            template = template.replace("{", "")
            template = template.replace("}", "")
            opt = []
            for f in fnames:
                opt.append(template.replace(fname, f))

            result = "={" + ",".join(opt) + "}"

    return result



""" This is the RFIO StorageClass
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import re
import os
import time

import six

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Resources.Storage.Utilities import checkArgumentFormat
from DIRAC.Resources.Storage.StorageBase import StorageBase
from DIRAC.Core.Utilities.Subprocess import shellCall
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Utilities.File import getSize


class RFIOStorage(StorageBase):

    _INPUT_PROTOCOLS = ["file", "rfio"]
    _OUTPUT_PROTOCOLS = ["rfio"]

    def __init__(self, storageName, parameters):

        StorageBase.__init__(self, storageName, parameters)
        self.spaceToken = self.protocolParameters["SpaceToken"]

        self.isok = True

        self.pluginName = "RFIO"

        self.timeout = 100
        self.long_timeout = 600

    #############################################################
    #
    # These are the methods for manipulating the client
    #

    def getName(self):
        """The name with which the storage was instantiated"""
        return S_OK(self.name)

    #############################################################
    #
    # These are the methods for file manipulation
    #

    def exists(self, path):
        """Check if the given path exists. The 'path' variable can be a string or a list of strings."""
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        gLogger.debug("RFIOStorage.exists: Determining the existance of %s files." % len(urls))
        comm = "nsls -d"
        for url in urls:
            comm = " %s %s" % (comm, url)
        res = shellCall(self.timeout, comm)
        successful = {}
        failed = {}
        if res["OK"]:
            returncode, stdout, stderr = res["Value"]
            if returncode in [0, 1]:
                for line in stdout.splitlines():
                    url = line.strip()
                    successful[url] = True
                for line in stderr.splitlines():
                    pfn, _ = line.split(": ")
                    url = pfn.strip()
                    successful[url] = False
            else:
                errStr = "RFIOStorage.exists: Completely failed to determine the existance files."
                gLogger.error(errStr, "%s %s" % (self.name, stderr))
                return S_ERROR(errStr)
        else:
            errStr = "RFIOStorage.exists: Completely failed to determine the existance files."
            gLogger.error(errStr, "%s %s" % (self.name, res["Message"]))
            return S_ERROR(errStr)
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def isFile(self, path):
        """Check if the given path exists and it is a file"""
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        gLogger.debug("RFIOStorage.isFile: Determining whether %s paths are files." % len(urls))
        successful = {}
        failed = {}
        comm = "nsls -ld"
        for url in urls:
            comm = " %s %s" % (comm, url)
        res = shellCall(self.timeout, comm)
        if not res["OK"]:
            return res
        returncode, stdout, stderr = res["Value"]
        if returncode in [0, 1]:
            for line in stdout.splitlines():
                permissions, _subdirs, _owner, _group, _size, _month, _date, _timeYear, pfn = line.split()
                if permissions[0] != "d":
                    successful[pfn] = True
                else:
                    successful[pfn] = False
            for line in stderr.splitlines():
                pfn, error = line.split(": ")
                url = pfn.strip()
                failed[url] = error
        else:
            errStr = "RFIOStorage.isFile: Completely failed to determine whether path is file."
            gLogger.error(errStr, "%s %s" % (self.name, stderr))
            return S_ERROR(errStr)
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def __getPathMetadata(self, urls):
        gLogger.debug("RFIOStorage.__getPathMetadata: Attempting to get metadata for %s paths." % (len(urls)))
        comm = "nsls -ld"
        for url in urls:
            comm = " %s %s" % (comm, url)
        res = shellCall(self.timeout, comm)
        successful = {}
        failed = {}
        if not res["OK"]:
            errStr = "RFIOStorage.__getPathMetadata: Completely failed to get path metadata."
            gLogger.error(errStr, res["Message"])
            return S_ERROR(errStr)
        else:
            returncode, stdout, stderr = res["Value"]
            if returncode not in [0, 1]:
                errStr = "RFIOStorage.__getPathMetadata: failed to perform nsls."
                gLogger.error(errStr, stderr)
            else:
                for line in stdout.splitlines():
                    permissions, subdirs, owner, group, size, month, date, timeYear, pfn = line.split()
                    successful[pfn] = {}
                    if permissions[0] == "d":
                        successful[pfn]["Type"] = "Directory"
                    else:
                        successful[pfn]["Type"] = "File"
                    successful[pfn]["Mode"] = self.__permissionsToInt(permissions)
                    successful[pfn]["NbSubDirs"] = subdirs
                    successful[pfn]["Owner"] = owner
                    successful[pfn]["Group"] = group
                    successful[pfn]["Size"] = int(size)
                    successful[pfn]["Month"] = month
                    successful[pfn]["Date"] = date
                    successful[pfn]["Year"] = timeYear
                for line in stderr.splitlines():
                    pfn, error = line.split(": ")
                    url = pfn.strip()
                    failed[url] = error
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def __permissionsToInt(self, permissions):
        mode = permissions[1:]
        return sum(pow(2, 8 - i) * int(mode[i] != "-") for i in range(0, 9))

    def __getFileMetadata(self, urls):
        gLogger.debug(
            "RFIOStorage.__getPathMetadata: Attempting to get additional metadata for %s files." % (len(urls))
        )
        # Check whether the files that exist are staged
        comm = "stager_qry -S %s" % self.spaceToken
        successful = {}
        for pfn in urls:
            successful[pfn] = {}
            comm = "%s -M %s" % (comm, pfn)
        res = shellCall(self.timeout, comm)
        if not res["OK"]:
            errStr = "RFIOStorage.__getFileMetadata: Completely failed to get cached status."
            gLogger.error(errStr, res["Message"])
            return S_ERROR(errStr)
        else:
            _returncode, stdout, _stderr = res["Value"]
            for line in stdout.splitlines():
                pfn = line.split()[0]
                status = line.split()[-1]
                if status in ["STAGED", "CANBEMIGR"]:
                    successful[pfn]["Cached"] = True
        for pfn in urls:
            if "Cached" not in successful[pfn]:
                successful[pfn]["Cached"] = False

        # Now for the files that exist get the tape segment (i.e. whether they have been migrated) and related checksum
        comm = "nsls -lT --checksum"
        for pfn in urls:
            comm = "%s %s" % (comm, pfn)
        res = shellCall(self.timeout, comm)
        if not res["OK"]:
            errStr = "RFIOStorage.__getFileMetadata: Completely failed to get migration status."
            gLogger.error(errStr, res["Message"])
            return S_ERROR(errStr)
        else:
            _returncode, stdout, _stderr = res["Value"]
            for line in stdout.splitlines():
                pfn = line.split()[-1]
                checksum = line.split()[-2]
                successful[pfn]["Migrated"] = True
                successful[pfn]["Checksum"] = checksum
        for pfn in urls:
            if "Migrated" not in successful[pfn]:
                successful[pfn]["Migrated"] = False

        # Update all the metadata with the common one
        for lfn in successful:
            successful[lfn] = self._addCommonMetadata(successful[lfn])

        resDict = {"Failed": {}, "Successful": successful}
        return S_OK(resDict)

    def getFile(self, path, localPath=False):
        """Get a local copy in the current directory of a physical file specified by its path"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        failed = {}
        successful = {}
        for src_url in urls.keys():
            fileName = os.path.basename(src_url)
            if localPath:
                dest_file = "%s/%s" % (localPath, fileName)
            else:
                dest_file = "%s/%s" % (os.getcwd(), fileName)
            res = self.__getFile(src_url, dest_file)
            if res["OK"]:
                successful[src_url] = res["Value"]
            else:
                failed[src_url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def __getFile(self, src_url, dest_file):
        """Get a local copy in the current directory of a physical file specified by its path"""
        if not os.path.exists(os.path.dirname(dest_file)):
            os.makedirs(os.path.dirname(dest_file))
        if os.path.exists(dest_file):
            gLogger.debug("RFIOStorage.getFile: Local file already exists %s. Removing..." % dest_file)
            os.remove(dest_file)
        res = self.__executeOperation(src_url, "getFileSize")
        if not res["OK"]:
            return S_ERROR(res["Message"])
        remoteSize = res["Value"]
        MIN_BANDWIDTH = 1024 * 100  # 100 KB/s
        timeout = int(remoteSize / MIN_BANDWIDTH + 300)
        gLogger.debug("RFIOStorage.getFile: Executing transfer of %s to %s" % (src_url, dest_file))
        comm = "rfcp %s %s" % (src_url, dest_file)
        res = shellCall(timeout, comm)
        if res["OK"]:
            returncode, _stdout, stderr = res["Value"]
            if returncode == 0:
                gLogger.debug("RFIOStorage.__getFile: Got file from storage, performing post transfer check.")
                localSize = getSize(dest_file)
                if localSize == remoteSize:
                    gLogger.debug("RFIOStorage.getFile: Post transfer check successful.")
                    return S_OK(localSize)
                errorMessage = "RFIOStorage.__getFile: Source and destination file sizes do not match."
                gLogger.error(errorMessage, src_url)
            else:
                errStr = "RFIOStorage.__getFile: Failed to get local copy of file."
                gLogger.error(errStr, stderr)
                errorMessage = "%s %s" % (errStr, stderr)
        else:
            errStr = "RFIOStorage.__getFile: Failed to get local copy of file."
            gLogger.error(errStr, res["Message"])
            errorMessage = "%s %s" % (errStr, res["Message"])
        if os.path.exists(dest_file):
            gLogger.debug("RFIOStorage.getFile: Removing local file %s." % dest_file)
            os.remove(dest_file)
        return S_ERROR(errorMessage)

    def putFile(self, path, sourceSize=0):
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        failed = {}
        successful = {}
        for dest_url, src_file in urls.items():
            res = self.__executeOperation(os.path.dirname(dest_url), "createDirectory")
            if not res["OK"]:
                failed[dest_url] = res["Message"]
            else:
                res = self.__putFile(src_file, dest_url, sourceSize)
                if res["OK"]:
                    successful[dest_url] = res["Value"]
                else:
                    failed[dest_url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def __putFile(self, src_file, dest_url, sourceSize):
        """Put a copy of the local file to the current directory on the physical storage"""
        # Pre-transfer check
        res = self.__executeOperation(dest_url, "exists")
        if not res["OK"]:
            gLogger.debug("RFIOStorage.__putFile: Failed to find pre-existance of destination file.")
            return res
        if res["Value"]:
            res = self.__executeOperation(dest_url, "removeFile")
            if not res["OK"]:
                gLogger.debug("RFIOStorage.__putFile: Failed to remove remote file %s." % dest_url)
            else:
                gLogger.debug("RFIOStorage.__putFile: Removed remote file %s." % dest_url)
        if not os.path.exists(src_file):
            errStr = "RFIOStorage.__putFile: The source local file does not exist."
            gLogger.error(errStr, src_file)
            return S_ERROR(errStr)
        sourceSize = getSize(src_file)
        if sourceSize == -1:
            errStr = "RFIOStorage.__putFile: Failed to get file size."
            gLogger.error(errStr, src_file)
            return S_ERROR(errStr)

        res = self.__getTransportURL(dest_url)
        if not res["OK"]:
            gLogger.debug("RFIOStorage.__putFile: Failed to get transport URL for file.")
            return res
        turl = res["Value"]

        MIN_BANDWIDTH = 1024 * 100  # 100 KB/s
        timeout = sourceSize / MIN_BANDWIDTH + 300
        gLogger.debug("RFIOStorage.putFile: Executing transfer of %s to %s" % (src_file, turl))
        comm = "rfcp %s '%s'" % (src_file, turl)
        res = shellCall(timeout, comm)
        if res["OK"]:
            returncode, _stdout, stderr = res["Value"]
            if returncode == 0:
                gLogger.debug("RFIOStorage.putFile: Put file to storage, performing post transfer check.")
                res = self.__executeOperation(dest_url, "getFileSize")
                if res["OK"]:
                    destinationSize = res["Value"]
                    if sourceSize == destinationSize:
                        gLogger.debug("RFIOStorage.__putFile: Post transfer check successful.")
                        return S_OK(destinationSize)
                errorMessage = "RFIOStorage.__putFile: Source and destination file sizes do not match."
                gLogger.error(errorMessage, dest_url)
            else:
                errStr = "RFIOStorage.__putFile: Failed to put file to remote storage."
                gLogger.error(errStr, stderr)
                errorMessage = "%s %s" % (errStr, stderr)
        else:
            errStr = "RFIOStorage.__putFile: Failed to put file to remote storage."
            gLogger.error(errStr, res["Message"])
            errorMessage = "%s %s" % (errStr, res["Message"])
        res = self.__executeOperation(dest_url, "removeFile")
        if res["OK"]:
            gLogger.debug("RFIOStorage.__putFile: Removed remote file remnant %s." % dest_url)
        else:
            gLogger.debug("RFIOStorage.__putFile: Unable to remove remote file remnant %s." % dest_url)
        return S_ERROR(errorMessage)

    def removeFile(self, path):
        """Remove physically the file specified by its path"""
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        listOfLists = breakListIntoChunks(urls, 100)
        for urls in listOfLists:
            gLogger.debug("RFIOStorage.removeFile: Attempting to remove %s files." % len(urls))
            comm = "stager_rm -S %s" % self.spaceToken
            for url in urls:
                comm = "%s -M %s" % (comm, url)
            res = shellCall(100, comm)
            if res["OK"]:
                returncode, _stdout, stderr = res["Value"]
                if returncode in [0, 1]:
                    comm = "nsrm -f"
                    for url in urls:
                        comm = "%s %s" % (comm, url)
                    res = shellCall(100, comm)
                    if res["OK"]:
                        returncode, _stdout, stderr = res["Value"]
                        if returncode in [0, 1]:
                            for pfn in urls:
                                successful[pfn] = True
                        else:
                            errStr = "RFIOStorage.removeFile. Completely failed to remove files from the nameserver."
                            gLogger.error(errStr, stderr)
                            for pfn in urls:
                                failed[pfn] = errStr
                    else:
                        errStr = "RFIOStorage.removeFile. Completely failed to remove files from the nameserver."
                        gLogger.error(errStr, res["Message"])
                        for pfn in urls:
                            failed[pfn] = errStr
                else:
                    errStr = "RFIOStorage.removeFile. Completely failed to remove files from the stager."
                    gLogger.error(errStr, stderr)
                    for pfn in urls:
                        failed[pfn] = errStr
            else:
                errStr = "RFIOStorage.removeFile. Completely failed to remove files from the stager."
                gLogger.error(errStr, res["Message"])
                for pfn in urls:
                    failed[pfn] = errStr
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getFileMetadata(self, path):
        """Get metadata associated to the file"""
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        gLogger.debug("RFIOStorage.getFileMetadata: Obtaining metadata for %s files." % len(urls))
        res = self.__getPathMetadata(urls)
        if not res["OK"]:
            return res
        failed = {}
        successful = {}
        for pfn, error in res["Value"]["Failed"].items():
            if error == "No such file or directory":
                failed[pfn] = "File does not exist"
            else:
                failed[pfn] = error
        files = []
        for pfn, pfnDict in res["Value"]["Successful"].items():
            if pfnDict["Type"] == "Directory":
                failed[pfn] = "Supplied path is not a file"
            else:
                successful[pfn] = res["Value"]["Successful"][pfn]
                files.append(pfn)
        if files:
            res = self.__getFileMetadata(files)
            if not res["OK"]:
                return res
            for pfn, pfnDict in res["Value"]["Successful"].items():
                successful[pfn].update(pfnDict)
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getFileSize(self, path):
        """Get the physical size of the given file"""
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        gLogger.debug("RFIOStorage.getFileSize: Determining the sizes for  %s files." % len(urls))
        res = self.__getPathMetadata(urls)
        if not res["OK"]:
            return res
        failed = {}
        successful = {}
        for pfn, error in res["Value"]["Failed"].items():
            if error == "No such file or directory":
                failed[pfn] = "File does not exist"
            else:
                failed[pfn] = error
        for pfn, pfnDict in res["Value"]["Successful"].items():
            if pfnDict["Type"] == "Directory":
                failed[pfn] = "Supplied path is not a file"
            else:
                successful[pfn] = res["Value"]["Successful"][pfn]["Size"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def prestageFile(self, path):
        """Issue prestage request for file"""
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        userTag = "%s-%s" % (self.spaceToken, time.time())
        comm = "stager_get -S %s -U %s " % (self.spaceToken, userTag)
        for url in urls:
            comm = "%s -M %s" % (comm, url)
        res = shellCall(100, comm)
        successful = {}
        failed = {}
        if res["OK"]:
            returncode, stdout, stderr = res["Value"]
            if returncode in [0, 1]:
                for line in stdout.splitlines():
                    if re.search("SUBREQUEST_READY", line):
                        pfn, _status = line.split()
                        successful[pfn] = userTag
                    elif re.search("SUBREQUEST_FAILED", line):
                        pfn, _status, err = line.split(" ", 2)
                        failed[pfn] = err
            else:
                errStr = "RFIOStorage.prestageFile: Got unexpected return code from stager_get."
                gLogger.error(errStr, stderr)
                return S_ERROR(errStr)
        else:
            errStr = "RFIOStorage.prestageFile: Completely failed to issue stage requests."
            gLogger.error(errStr, res["Message"])
            return S_ERROR(errStr)
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def prestageFileStatus(self, path):
        """Monitor the status of a prestage request"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        requestFiles = {}
        for url, requestID in urls.items():
            if requestID not in requestFiles:
                requestFiles[requestID] = []
            requestFiles[requestID].append(url)
        for requestID, urls in requestFiles.items():
            comm = "stager_qry -S %s -U %s " % (self.spaceToken, requestID)
            res = shellCall(100, comm)
            if res["OK"]:
                returncode, stdout, stderr = res["Value"]
                if returncode in [0, 1]:
                    for line in stdout.splitlines():
                        pfn = line.split()[0]
                        status = line.split()[-1]
                        if status in ["STAGED", "CANBEMIGR"]:
                            successful[pfn] = True
                        else:
                            successful[pfn] = False
                else:
                    errStr = "RFIOStorage.prestageFileStatus: Got unexpected return code from stager_get."
                    gLogger.error(errStr, stderr)
                    return S_ERROR(errStr)
            else:
                errStr = "RFIOStorage.prestageFileStatus: Completely failed to obtain prestage status."
                gLogger.error(errStr, res["Message"])
                return S_ERROR(errStr)
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getTransportURL(self, path, protocols=False):
        """Obtain the TURLs for the supplied path and protocols"""
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        res = self.exists(urls)
        if not res["OK"]:
            return res
        for path, exists in res["Value"]["Successful"].items():
            if not exists:
                failed[path] = "File does not exist"
            else:
                res = self.__getTransportURL(path)
                if not res["OK"]:
                    failed[path] = res["Message"]
                else:
                    successful[path] = res["Value"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def __getTransportURL(self, path):
        try:
            if self.spaceToken:
                tURL = "%s://%s:%s/?svcClass=%s&castorVersion=2&path=%s" % (
                    self.protocolParameters["Protocol"],
                    self.protocolParameters["Host"],
                    self.protocolParameters["Port"],
                    self.spaceToken,
                    path,
                )
            else:
                tURL = "castor:%s" % (path)
            return S_OK(tURL)
        except Exception as x:
            errStr = "RFIOStorage.__getTransportURL: Exception while creating turl."
            gLogger.exception(errStr, self.name, x)
            return S_ERROR(errStr)

    #############################################################
    #
    # These are the methods for directory manipulation
    #

    def isDirectory(self, path):
        """Check if the given path exists and it is a directory"""
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        gLogger.debug("RFIOStorage.isDirectory: Determining whether %s paths are directories." % len(urls))
        res = self.__getPathMetadata(urls)
        if not res["OK"]:
            return res
        failed = {}
        successful = {}
        for pfn, error in res["Value"]["Failed"].items():
            if error == "No such file or directory":
                failed[pfn] = "Directory does not exist"
            else:
                failed[pfn] = error
        for pfn, pfnDict in res["Value"]["Successful"].items():
            if pfnDict["Type"] == "Directory":
                successful[pfn] = True
            else:
                successful[pfn] = False
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getDirectory(self, path, localPath=False):
        """Get locally a directory from the physical storage together with all its files and subdirectories."""
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        successful = {}
        failed = {}
        gLogger.debug("RFIOStorage.getDirectory: Attempting to get local copies of %s directories." % len(urls))
        for src_directory in urls:
            dirName = os.path.basename(src_directory)
            if localPath:
                dest_dir = "%s/%s" % (localPath, dirName)
            else:
                dest_dir = "%s/%s" % (os.getcwd(), dirName)
            res = self.__getDir(src_directory, dest_dir)
            if res["OK"]:
                if res["Value"]["AllGot"]:
                    gLogger.debug("RFIOStorage.getDirectory: Successfully got local copy of %s" % src_directory)
                    successful[src_directory] = {"Files": res["Value"]["Files"], "Size": res["Value"]["Size"]}
                else:
                    gLogger.error("RFIOStorage.getDirectory: Failed to get entire directory.", src_directory)
                    failed[src_directory] = {"Files": res["Value"]["Files"], "Size": res["Value"]["Size"]}
            else:
                gLogger.error(
                    "RFIOStorage.getDirectory: Completely failed to get local copy of directory.", src_directory
                )
                failed[src_directory] = {"Files": 0, "Size": 0}
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def __getDir(self, srcDirectory, destDirectory):
        """Black magic contained within..."""
        filesGot = 0
        sizeGot = 0

        # Check the remote directory exists
        res = self.isDirectory(srcDirectory)
        if not res["OK"]:
            errStr = "RFIOStorage.__getDir: Failed to find the supplied source directory."
            gLogger.error(errStr, srcDirectory)
            return S_ERROR(errStr)
        if srcDirectory not in res["Value"]["Successful"]:
            errStr = "RFIOStorage.__getDir: Failed to find the supplied source directory."
            gLogger.error(errStr, srcDirectory)
            return S_ERROR(errStr)
        if not res["Value"]["Successful"][srcDirectory]:
            errStr = "RFIOStorage.__getDir: The supplied source directory does not exist."
            gLogger.error(errStr, srcDirectory)
            return S_ERROR(errStr)

        # Check the local directory exists and create it if not
        if not os.path.exists(destDirectory):
            os.makedirs(destDirectory)

        # Get the remote directory contents
        res = self.listDirectory(srcDirectory)
        if not res["OK"]:
            errStr = "RFIOStorage.__getDir: Failed to list the source directory."
            gLogger.error(errStr, srcDirectory)
        if srcDirectory not in res["Value"]["Successful"]:
            errStr = "RFIOStorage.__getDir: Failed to list the source directory."
            gLogger.error(errStr, srcDirectory)

        surlsDict = res["Value"]["Successful"][srcDirectory]["Files"]
        subDirsDict = res["Value"]["Successful"][srcDirectory]["SubDirs"]

        # First get all the files in the directory
        gotFiles = True
        for surl in surlsDict.keys():
            surlGot = False
            fileSize = surlsDict[surl]["Size"]
            fileName = os.path.basename(surl)
            localPath = "%s/%s" % (destDirectory, fileName)
            fileDict = {surl: localPath}
            res = self.getFile(fileDict)
            if res["OK"]:
                if surl in res["Value"]["Successful"]:
                    filesGot += 1
                    sizeGot += fileSize
                    surlGot = True
            if not surlGot:
                gotFiles = False

        # Then recursively get the sub directories
        subDirsGot = True
        for subDir in subDirsDict.keys():
            subDirName = os.path.basename(subDir)
            localPath = "%s/%s" % (destDirectory, subDirName)
            dirSuccessful = False
            res = self.__getDir(subDir, localPath)
            if res["OK"]:
                if res["Value"]["AllGot"]:
                    dirSuccessful = True
                filesGot += res["Value"]["Files"]
                sizeGot += res["Value"]["Size"]
            if not dirSuccessful:
                subDirsGot = False

        # Check whether all the operations were successful
        if subDirsGot and gotFiles:
            allGot = True
        else:
            allGot = False
        resDict = {"AllGot": allGot, "Files": filesGot, "Size": sizeGot}
        return S_OK(resDict)

    def putDirectory(self, path):
        """Put a local directory to the physical storage together with all its files and subdirectories."""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        successful = {}
        failed = {}
        gLogger.debug("RFIOStorage.putDirectory: Attemping to put %s directories to remote storage." % len(urls))
        for destDir, sourceDir in urls.items():
            res = self.__putDir(sourceDir, destDir)
            if res["OK"]:
                if res["Value"]["AllPut"]:
                    gLogger.debug(
                        "RFIOStorage.putDirectory: Successfully put directory to remote storage: %s" % destDir
                    )
                    successful[destDir] = {"Files": res["Value"]["Files"], "Size": res["Value"]["Size"]}
                else:
                    gLogger.error(
                        "RFIOStorage.putDirectory: Failed to put entire directory to remote storage.", destDir
                    )
                    failed[destDir] = {"Files": res["Value"]["Files"], "Size": res["Value"]["Size"]}
            else:
                gLogger.error(
                    "RFIOStorage.putDirectory: Completely failed to put directory to remote storage.", destDir
                )
                failed[destDir] = {"Files": 0, "Size": 0}
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def __putDir(self, src_directory, dest_directory):
        """Black magic contained within..."""
        filesPut = 0
        sizePut = 0

        # Check the local directory exists
        if not os.path.isdir(src_directory):
            errStr = "RFIOStorage.__putDir: The supplied source directory does not exist."
            gLogger.error(errStr, src_directory)
            return S_ERROR(errStr)

        # Create the remote directory
        res = self.createDirectory(dest_directory)
        if not res["OK"]:
            errStr = "RFIOStorage.__putDir: Failed to create destination directory."
            gLogger.error(errStr, dest_directory)
            return S_ERROR(errStr)

        # Get the local directory contents
        contents = os.listdir(src_directory)
        allSuccessful = True
        for cFile in contents:
            pathSuccessful = False
            localPath = "%s/%s" % (src_directory, cFile)
            remotePath = "%s/%s" % (dest_directory, cFile)
            if os.path.isdir(localPath):
                res = self.__putDir(localPath, remotePath)
                if res["OK"]:
                    if res["Value"]["AllPut"]:
                        pathSuccessful = True
                    filesPut += res["Value"]["Files"]
                    sizePut += res["Value"]["Size"]
                else:
                    return S_ERROR("Failed to put directory")
            else:
                fileDict = {remotePath: localPath}
                res = self.putFile(fileDict)
                if res["OK"]:
                    if remotePath in res["Value"]["Successful"]:
                        filesPut += 1
                        sizePut += res["Value"]["Successful"][remotePath]
                        pathSuccessful = True
            if not pathSuccessful:
                allSuccessful = False
        resDict = {"AllPut": allSuccessful, "Files": filesPut, "Size": sizePut}
        return S_OK(resDict)

    def createDirectory(self, path):
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        gLogger.debug("RFIOStorage.createDirectory: Attempting to create %s directories." % len(urls))
        for url in urls:
            strippedUrl = url.rstrip("/")
            res = self.__makeDirs(strippedUrl)
            if res["OK"]:
                gLogger.debug("RFIOStorage.createDirectory: Successfully created directory on storage: %s" % url)
                successful[url] = True
            else:
                gLogger.error("RFIOStorage.createDirectory: Failed to create directory on storage.", url)
                failed[url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def __makeDir(self, path):
        # First create a local file that will be used as a directory place holder in storage name space
        comm = "nsmkdir -m 775 %s" % path
        res = shellCall(100, comm)
        if not res["OK"]:
            return res
        returncode, _stdout, stderr = res["Value"]
        if returncode not in [0]:
            return S_ERROR(stderr)
        return S_OK()

    def __makeDirs(self, path):
        """Black magic contained within...."""
        pDir = os.path.dirname(path)
        res = self.exists(path)
        if not res["OK"]:
            return res
        if res["OK"]:
            if path in res["Value"]["Successful"]:
                if res["Value"]["Successful"][path]:
                    return S_OK()
                else:
                    res = self.exists(pDir)
                    if res["OK"]:
                        if pDir in res["Value"]["Successful"]:
                            if res["Value"]["Successful"][pDir]:
                                res = self.__makeDir(path)
                            else:
                                res = self.__makeDirs(pDir)
                                res = self.__makeDir(path)
        return res

    def removeDirectory(self, path, recursive=False):
        """Remove a directory on the physical storage together with all its files and
        subdirectories.
        """
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        gLogger.debug("RFIOStorage.removeDirectory: Attempting to remove %s directories." % len(urls))
        successful = {}
        failed = {}
        for url in urls:
            comm = "nsrm -r %s" % url
            res = shellCall(100, comm)
            if res["OK"]:
                returncode, _stdout, stderr = res["Value"]
                if returncode == 0:
                    successful[url] = {"FilesRemoved": 0, "SizeRemoved": 0}
                elif returncode == 1:
                    successful[url] = {"FilesRemoved": 0, "SizeRemoved": 0}
                else:
                    failed[url] = stderr
            else:
                errStr = "RFIOStorage.removeDirectory: Completely failed to remove directory."
                gLogger.error(errStr, "%s %s" % (url, res["Message"]))
                failed[url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def listDirectory(self, path):
        """List the supplied path. First checks whether the path is a directory then gets the contents."""
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        gLogger.debug("RFIOStorage.listDirectory: Attempting to list %s directories." % len(urls))
        res = self.isDirectory(urls)
        if not res["OK"]:
            return res
        successful = {}
        failed = res["Value"]["Failed"]
        directories = []
        for url, isDirectory in res["Value"]["Successful"].items():
            if isDirectory:
                directories.append(url)
            else:
                errStr = "RFIOStorage.listDirectory: Directory does not exist."
                gLogger.error(errStr, url)
                failed[url] = errStr

        for directory in directories:
            comm = "nsls -l %s" % directory
            res = shellCall(self.timeout, comm)
            if res["OK"]:
                returncode, stdout, stderr = res["Value"]
                if not returncode == 0:
                    errStr = "RFIOStorage.listDirectory: Failed to list directory."
                    gLogger.error(errStr, "%s %s" % (directory, stderr))
                    failed[directory] = errStr
                else:
                    subDirs = {}
                    files = {}
                    successful[directory] = {}
                    for line in stdout.splitlines():
                        permissions, _subdirs, _owner, _group, size, _month, _date, _timeYear, pfn = line.split()
                        if not pfn == "dirac_directory":
                            path = "%s/%s" % (directory, pfn)
                            if permissions[0] == "d":
                                # If the subpath is a directory
                                subDirs[path] = True
                            elif permissions[0] == "m":
                                # In the case that the path is a migrated file
                                files[path] = {"Size": int(size), "Migrated": 1}
                            else:
                                # In the case that the path is not migrated file
                                files[path] = {"Size": int(size), "Migrated": 0}
                    successful[directory]["SubDirs"] = subDirs
                    successful[directory]["Files"] = files
            else:
                errStr = "RFIOStorage.listDirectory: Completely failed to list directory."
                gLogger.error(errStr, "%s %s" % (directory, res["Message"]))
                return S_ERROR(errStr)
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getDirectoryMetadata(self, path):
        """Get the metadata for the directory"""
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        gLogger.debug("RFIOStorage.getDirectoryMetadata: Attempting to get metadata for %s directories." % len(urls))
        res = self.isDirectory(urls)
        if not res["OK"]:
            return res
        successful = {}
        failed = res["Value"]["Failed"]
        directories = []
        for url, isDirectory in res["Value"]["Successful"].items():
            if isDirectory:
                directories.append(url)
            else:
                errStr = "RFIOStorage.getDirectoryMetadata: Directory does not exist."
                gLogger.error(errStr, url)
                failed[url] = errStr
        res = self.__getPathMetadata(directories)
        if not res["OK"]:
            return res
        else:
            failed.update(res["Value"]["Failed"])
            successful = res["Value"]["Successful"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getDirectorySize(self, path):
        """Get the size of the directory on the storage"""
        res = self.__checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        gLogger.debug("RFIOStorage.getDirectorySize: Attempting to get size of %s directories." % len(urls))
        res = self.listDirectory(urls)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        successful = {}
        for directory, dirDict in res["Value"]["Successful"].items():
            directorySize = 0
            directoryFiles = 0
            filesDict = dirDict["Files"]
            for _fileURL, fileDict in filesDict.items():
                directorySize += fileDict["Size"]
                directoryFiles += 1
            gLogger.debug("RFIOStorage.getDirectorySize: Successfully obtained size of %s." % directory)
            successful[directory] = {"Files": directoryFiles, "Size": directorySize}
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def __checkArgumentFormat(self, path):
        """FIXME: Can be replaced by a generic checkArgumentFormat Utility"""
        if isinstance(path, six.string_types):
            urls = [path]
        elif isinstance(path, list):
            urls = path
        elif isinstance(path, dict):
            urls = list(path)
        else:
            return S_ERROR("RFIOStorage.__checkArgumentFormat: Supplied path is not of the correct format.")
        return S_OK(urls)

    def __executeOperation(self, url, method):
        """Executes the requested functionality with the supplied url"""
        fcn = None
        if hasattr(self, method) and callable(getattr(self, method)):
            fcn = getattr(self, method)
        if not fcn:
            return S_ERROR("Unable to invoke %s, it isn't a member funtion of RFIOStorage" % method)
        res = fcn(url)
        if not res["OK"]:
            return res
        elif url not in res["Value"]["Successful"]:
            return S_ERROR(res["Value"]["Failed"][url])
        return S_OK(res["Value"]["Successful"][url])



""" This contains a bunch of tests for the RMS that can be ran
either as unit test for the DB (Test_RequestDB.py) or as
integration tests (Test_ReqDB.py)
"""

# pylint: disable=invalid-name,wrong-import-position

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import time


from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File

from DIRAC.RequestManagementSystem.DB import RequestDB

STRESS_REQUESTS = 10
BULK_REQUESTS = 10


def test_stress(reqDB):
    """stress test"""

    reqIDs = []
    for i in range(STRESS_REQUESTS):
        request = Request({"RequestName": "test-%d" % i})
        op = Operation({"Type": "RemoveReplica", "TargetSE": "CERN-USER"})
        op += File({"LFN": "/lhcb/user/c/cibak/foo"})
        request += op
        put = reqDB.putRequest(request)
        assert put["OK"], put
        reqIDs.append(put["Value"])

    startTime = time.time()

    for reqID in reqIDs:
        get = reqDB.getRequest(reqID)
        assert get["OK"], get

    endTime = time.time()

    print("getRequest duration %s " % (endTime - startTime))
    for reqID in reqIDs:
        delete = reqDB.deleteRequest(reqID)
        assert delete["OK"], delete


def test_stressBulk(reqDB):
    """stress test bulk"""

    reqIDs = []
    for i in range(STRESS_REQUESTS):
        request = Request({"RequestName": "test-%d" % i})
        op = Operation({"Type": "RemoveReplica", "TargetSE": "CERN-USER"})
        op += File({"LFN": "/lhcb/user/c/cibak/foo"})
        request += op
        put = reqDB.putRequest(request)
        assert put["OK"], put
        reqIDs.append(put["Value"])

    loops = STRESS_REQUESTS // BULK_REQUESTS + (1 if (STRESS_REQUESTS % BULK_REQUESTS) else 0)
    totalSuccessful = 0

    time.sleep(1)
    startTime = time.time()

    for i in range(loops):
        get = reqDB.getBulkRequests(BULK_REQUESTS, True)
        assert get["OK"], get

        totalSuccessful += len(get["Value"])

    endTime = time.time()

    print("getRequests duration %s " % (endTime - startTime))

    assert totalSuccessful == STRESS_REQUESTS, "Did not retrieve all the requests: %s instead of %s" % (
        totalSuccessful,
        STRESS_REQUESTS,
    )

    for reqID in reqIDs:
        delete = reqDB.deleteRequest(reqID)
        assert delete["OK"], delete


def test_scheduled(reqDB):
    """scheduled request r/w"""

    req = Request({"RequestName": "FTSTest"})
    op = Operation({"Type": "ReplicateAndRegister", "TargetSE": "CERN-USER"})
    op += File({"LFN": "/a/b/c", "Status": "Scheduled", "Checksum": "123456", "ChecksumType": "ADLER32"})
    req += op

    put = reqDB.putRequest(req)
    assert put["OK"], put
    reqID = put["Value"]

    peek = reqDB.peekRequest(reqID)
    assert peek["OK"], peek

    peek = peek["Value"]
    for op in peek:
        opId = op.OperationID

    getFTS = reqDB.getScheduledRequest(opId)
    assert getFTS["OK"], getFTS
    assert getFTS["Value"].RequestName == "FTSTest", "Wrong request name %s" % getFTS["Value"].RequestName

    delete = reqDB.deleteRequest(reqID)
    assert delete["OK"], delete


def test_dirty(reqDB):
    """dirty records

    This illustrates this bug https://github.com/sqlalchemy/sqlalchemy/discussions/6294
    """
    req = Request()
    req.RequestName = "dirty"

    op1 = Operation({"Type": "ReplicateAndRegister", "TargetSE": "CERN-USER"})
    op1 += File({"LFN": "/a/b/c/1", "Status": "Scheduled", "Checksum": "123456", "ChecksumType": "ADLER32"})

    op2 = Operation({"Type": "ReplicateAndRegister", "TargetSE": "CNAF-USER"})
    op2 += File({"LFN": "/a/b/c/2", "Status": "Scheduled", "Checksum": "123456", "ChecksumType": "ADLER32"})

    op3 = Operation({"Type": "ReplicateAndRegister", "TargetSE": "IN2P3-USER"})
    op3 += File({"LFN": "/a/b/c/3", "Status": "Scheduled", "Checksum": "123456", "ChecksumType": "ADLER32"})

    req += op1
    req += op2
    req += op3

    put = reqDB.putRequest(req)
    assert put["OK"], put
    reqID = put["Value"]

    get = reqDB.getRequest(reqID)
    assert get["OK"], get
    req = get["Value"]

    del req[0]
    assert len(req) == 2, "Wrong number of operations (%s) %s" % (len(req), req)

    put = reqDB.putRequest(req)
    assert put["OK"], put
    reqID = put["Value"]

    get = reqDB.getRequest(reqID)
    assert get["OK"], get
    req = get["Value"]

    assert len(req) == 2, "Wrong number of operations (%s) %s" % (len(req), req)

    op4 = Operation({"Type": "ReplicateAndRegister", "TargetSE": "CERN-USER"})
    op4 += File({"LFN": "/a/b/c/4", "Status": "Scheduled", "Checksum": "123456", "ChecksumType": "ADLER32"})

    req[0] = op4
    put = reqDB.putRequest(req)
    assert put["OK"], put
    reqID = put["Value"]

    get = reqDB.getRequest(reqID)
    assert get["OK"], get
    req = get["Value"]

    assert len(req) == 2, "Wrong number of operations (%s) %s" % (len(req), req)

    delete = reqDB.deleteRequest(reqID)
    assert delete["OK"], delete



""" Base class for MyProxy and VOMS
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import tempfile

import DIRAC
from DIRAC import gConfig, S_OK, S_ERROR
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.Core.Security import Locations


class BaseSecurity(object):
    def __init__(self, server=False, serverCert=False, serverKey=False, timeout=False):
        if timeout:
            self._secCmdTimeout = timeout
        else:
            self._secCmdTimeout = 30
        if not server:
            self._secServer = gConfig.getValue("/DIRAC/VOPolicy/MyProxyServer", "myproxy.cern.ch")
        else:
            self._secServer = server
        ckLoc = Locations.getHostCertificateAndKeyLocation()
        if serverCert:
            self._secCertLoc = serverCert
        else:
            if ckLoc:
                self._secCertLoc = ckLoc[0]
            else:
                self._secCertLoc = "%s/etc/grid-security/servercert.pem" % DIRAC.rootPath
        if serverKey:
            self._secKeyLoc = serverKey
        else:
            if ckLoc:
                self._secKeyLoc = ckLoc[1]
            else:
                self._secKeyLoc = "%s/etc/grid-security/serverkey.pem" % DIRAC.rootPath
        self._secRunningFromTrustedHost = gConfig.getValue("/DIRAC/VOPolicy/MyProxyTrustedHost", "True").lower() in (
            "y",
            "yes",
            "true",
        )
        self._secMaxProxyHours = gConfig.getValue("/DIRAC/VOPolicy/MyProxyMaxDelegationTime", 168)

    def getMyProxyServer(self):
        return self._secServer

    def getServiceDN(self):
        chain = X509Chain()
        retVal = chain.loadChainFromFile(self._secCertLoc)
        if not retVal["OK"]:
            return retVal
        return chain.getCertInChain(0)["Value"].getSubjectDN()

    def _getExternalCmdEnvironment(self):
        return dict(os.environ)

    def _unlinkFiles(self, files):
        if isinstance(files, (list, tuple)):
            for fileName in files:
                self._unlinkFiles(fileName)
        else:
            try:
                os.unlink(files)
            except Exception:
                pass

    def _generateTemporalFile(self):
        try:
            fd, filename = tempfile.mkstemp()
            os.close(fd)
        except IOError:
            return S_ERROR(DErrno.ECTMPF)
        return S_OK(filename)

    def _getUsername(self, proxyChain):
        retVal = proxyChain.getCredentials()
        if not retVal["OK"]:
            return retVal
        credDict = retVal["Value"]
        if not credDict["isProxy"]:
            return S_ERROR(DErrno.EX509, "chain does not contain a proxy")
        if not credDict["validDN"]:
            return S_ERROR(DErrno.EDISET, "DN %s is not known in dirac" % credDict["subject"])
        if not credDict["validGroup"]:
            return S_ERROR(DErrno.EDISET, "Group %s is invalid for DN %s" % (credDict["group"], credDict["subject"]))
        mpUsername = "%s:%s" % (credDict["group"], credDict["username"])
        return S_OK(mpUsername)



#!/usr/bin/env python
"""
Stop DIRAC component using runsvctrl utility
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.disableCS()
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(
        " System:  Name of the system for the component (default *: all)", mandatory=False, default="*"
    )
    Script.registerArgument(
        (
            "Service: Name of the particular component (default *: all)",
            "Agent:   Name of the particular component (default *: all)",
        ),
        mandatory=False,
        default="*",
    )
    _, args = Script.parseCommandLine()
    system, component = Script.getPositionalArgs(group=True)

    from DIRAC.FrameworkSystem.Client.ComponentInstaller import gComponentInstaller

    __RCSID__ = "$Id$"

    if len(args) > 2:
        Script.showHelp(exitCode=1)

    if system != "*":
        if len(args) > 1:
            component = args[1]

    gComponentInstaller.exitOnError = True

    result = gComponentInstaller.runsvctrlComponent(system, component, "d")
    if not result["OK"]:
        print("ERROR:", result["Message"])
        exit(-1)

    gComponentInstaller.printStartupStatus(result["Value"])


if __name__ == "__main__":
    main()



""" :mod: FileTest
    =======================

    .. module: FileTest
    :synopsis: test cases for Files
    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    test cases for Files
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import pytest

from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File


def test_ctors():
    """File construction and (de)serialisation"""
    theFile = File()
    assert isinstance(theFile, File)

    fromDict = {
        "Size": 1,
        "LFN": "/test/lfn",
        "ChecksumType": "ADLER32",
        "Checksum": "123456",
        "Status": "Waiting",
    }
    try:
        theFile = File(fromDict)
    except AttributeError as error:
        print("AttributeError: %s" % str(error))

    assert isinstance(theFile, File)
    for key, value in fromDict.items():
        assert getattr(theFile, key) == value

    toJSON = theFile.toJSON()
    assert toJSON["OK"], "JSON serialization error"


def test_valid_properties():
    theFile = File()

    theFile.FileID = 1
    assert theFile.FileID == 1
    theFile.Status = "Done"
    assert theFile.Status == "Done"
    theFile.LFN = "/some/path/somewhere"
    assert theFile.LFN == "/some/path/somewhere"
    theFile.PFN = "/some/path/somewhere"
    assert theFile.PFN == "/some/path/somewhere"
    theFile.Attempt = 1
    assert theFile.Attempt == 1
    theFile.Size = 1
    assert theFile.Size == 1
    theFile.GUID = "2bbabe80-e2f1-11e1-9b23-0800200c9a66"
    assert theFile.GUID == "2bbabe80-e2f1-11e1-9b23-0800200c9a66"
    theFile.ChecksumType = "adler32"
    assert theFile.ChecksumType == "ADLER32"
    theFile.Checksum = "123456"
    assert theFile.Checksum == "123456"



""" PublisherHandler

This service has been built to provide the RSS web views with all the information
they need. NO OTHER COMPONENT THAN Web controllers should make use of it.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

#  pylint: disable=no-self-use

import six
from datetime import datetime, timedelta

# DIRAC
from DIRAC import S_OK, gConfig, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities.SiteSEMapping import getSEHosts, getStorageElementsHosts
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites, getSiteCEMapping
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.ResourceStatusSystem.Client.ResourceStatusClient import ResourceStatusClient
from DIRAC.ResourceStatusSystem.Utilities import Utils

ResourceManagementClient = getattr(
    Utils.voimport("DIRAC.ResourceStatusSystem.Client.ResourceManagementClient"), "ResourceManagementClient"
)


class PublisherHandler(RequestHandler):
    """
    RPCServer used to deliver data to the web portal.

    """

    def __init__(self, *args, **kwargs):
        """
        Constructor
        """
        super(PublisherHandler, self).__init__(*args, **kwargs)

    # ResourceStatusClient .......................................................

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """Initialization of client objects"""
        cls.rsClient = ResourceStatusClient()
        cls.rmClient = ResourceManagementClient()

        return S_OK()

    types_getSites = []

    def export_getSites(self):
        """
        Returns list of all sites considered by RSS

        :return: S_OK( [ sites ] ) | S_ERROR
        """

        return getSites()

    types_getSitesResources = [(six.string_types, list, type(None))]

    def export_getSitesResources(self, siteNames):
        """
        Returns dictionary with SEs and CEs for the given site(s). If siteNames is
        None, all sites are taken into account.

        :return: S_OK( { site1 : { ces : [ ces ], 'ses' : [ ses  ] },... } ) | S_ERROR
        """

        if siteNames is None:
            res = getSites()
            if not res["OK"]:
                self.log.error("Error getting sites", res["Message"])
                return res
            siteNames = res["Value"]

        if isinstance(siteNames, six.string_types):
            siteNames = [siteNames]

        sitesRes = {}
        for siteName in siteNames:

            result = getSiteCEMapping()
            if not result["OK"]:
                self.log.error("Error getting sites/CEs mapping", result["Message"])
                return result
            res = {}
            res["ces"] = result["Value"][siteName]
            # Convert StorageElements to host names
            result = DMSHelpers().getSiteSEMapping()
            if not result["OK"]:
                self.log.error("Error getting sites/SEs mapping", result["Message"])
                sitesRes[siteName] = res
                continue
            ses = result["Value"][1].get(siteName, [])
            result = getStorageElementsHosts(ses)
            if not result["OK"]:
                self.log.error("Error getting storage element hosts", result["Message"])
                return result
            # Remove duplicates
            res["ses"] = list(set(result["Value"]))

            sitesRes[siteName] = res

        return S_OK(sitesRes)

    types_getElementStatuses = [
        six.string_types,
        (six.string_types, list, type(None)),
        (six.string_types, list, type(None)),
        (six.string_types, list, type(None)),
        (six.string_types, list, type(None)),
        (six.string_types, list, type(None)),
    ]

    @classmethod
    def export_getElementStatuses(cls, element, name, elementType, statusType, status, tokenOwner):
        """
        Returns element statuses from the ResourceStatusDB
        """

        return cls.rsClient.selectStatusElement(
            element,
            "Status",
            name=name,
            elementType=elementType,
            statusType=statusType,
            status=status,
            tokenOwner=tokenOwner,
        )

    types_getElementHistory = [
        six.string_types,
        (six.string_types, list, type(None)),
        (six.string_types, list, type(None)),
        (six.string_types, list, type(None)),
    ]

    @classmethod
    def export_getElementHistory(cls, element, name, elementType, statusType):
        """
        Returns element history from ResourceStatusDB
        """

        columns = ["Status", "DateEffective", "Reason"]
        return cls.rsClient.selectStatusElement(
            element, "History", name=name, elementType=elementType, statusType=statusType, meta={"columns": columns}
        )

    types_getElementPolicies = [
        six.string_types,
        (six.string_types, list, type(None)),
        (six.string_types, list, type(None)),
    ]

    @classmethod
    def export_getElementPolicies(cls, element, name, statusType):
        """
        Returns policies for a given element
        """

        columns = ["Status", "PolicyName", "DateEffective", "LastCheckTime", "Reason"]
        return cls.rmClient.selectPolicyResult(
            element=element, name=name, statusType=statusType, meta={"columns": columns}
        )

    types_getNodeStatuses = []

    @classmethod
    def export_getNodeStatuses(cls):
        return cls.rsClient.selectStatusElement("Node", "Status")

    types_getTree = [six.string_types, six.string_types]

    def export_getTree(self, elementType, elementName):
        """
        Given an element type and name,
        finds its parent site and returns all descendants of that site.
        """

        site = self.getSite(elementType, elementName)
        if not site:
            return S_ERROR("No site")

        siteStatus = self.rsClient.selectStatusElement(
            "Site", "Status", name=site, meta={"columns": ["StatusType", "Status"]}
        )
        if not siteStatus["OK"]:
            return siteStatus

        tree = {site: {"statusTypes": dict(siteStatus["Value"])}}

        result = getSiteCEMapping()
        if not result["OK"]:
            return result
        ces = result["Value"][site]
        cesStatus = self.rsClient.selectStatusElement(
            "Resource", "Status", name=ces, meta={"columns": ["Name", "StatusType", "Status"]}
        )
        if not cesStatus["OK"]:
            return cesStatus

        res = DMSHelpers().getSiteSEMapping()
        if not res["OK"]:
            self.log.error("Could not get site to SE mapping", res["Message"])
            return S_OK()
        ses = res["Value"][1].get(site, [])
        sesStatus = self.rsClient.selectStatusElement(
            "Resource", "Status", name=list(ses), meta={"columns": ["Name", "StatusType", "Status"]}
        )
        if not sesStatus["OK"]:
            return sesStatus

        def feedTree(elementsList):

            elements = {}
            for elementTuple in elementsList["Value"]:
                name, statusType, status = elementTuple

                if name not in elements:
                    elements[name] = {}
                elements[name][statusType] = status

            return elements

        tree[site]["ces"] = feedTree(cesStatus)
        tree[site]["ses"] = feedTree(sesStatus)

        return S_OK(tree)

    types_setToken = [six.string_types] * 7

    def export_setToken(self, element, name, statusType, token, elementType, username, lastCheckTime):

        lastCheckTime = datetime.strptime(lastCheckTime, "%Y-%m-%d %H:%M:%S")

        credentials = self.getRemoteCredentials()
        self.log.info(credentials)

        elementInDB = self.rsClient.selectStatusElement(
            element, "Status", name=name, statusType=statusType, elementType=elementType, lastCheckTime=lastCheckTime
        )
        if not elementInDB["OK"]:
            return elementInDB
        elif not elementInDB["Value"]:
            return S_ERROR("Your selection has been modified. Please refresh.")

        if token == "Acquire":
            tokenOwner = username
            tokenExpiration = datetime.utcnow() + timedelta(days=1)
        elif token == "Release":
            tokenOwner = "rs_svc"
            tokenExpiration = datetime.max
        else:
            return S_ERROR("%s is unknown token action" % token)

        reason = "Token %sd by %s ( web )" % (token, username)

        newStatus = self.rsClient.addOrModifyStatusElement(
            element,
            "Status",
            name=name,
            statusType=statusType,
            elementType=elementType,
            reason=reason,
            tokenOwner=tokenOwner,
            tokenExpiration=tokenExpiration,
        )
        if not newStatus["OK"]:
            return newStatus

        return S_OK(reason)

    def getSite(self, elementType, elementName):
        """
        Given an element name, return its site
        """

        if elementType == "StorageElement":
            elementType = "SE"

        domainNames = gConfig.getSections("Resources/Sites")
        if not domainNames["OK"]:
            return domainNames
        domainNames = domainNames["Value"]

        for domainName in domainNames:

            sites = gConfig.getSections("Resources/Sites/%s" % domainName)
            if not sites["OK"]:
                continue

            for site in sites["Value"]:

                elements = gConfig.getValue("Resources/Sites/%s/%s/%s" % (domainName, site, elementType), "")
                if elementName in elements:
                    return site

        return ""

    # ResourceManagementClient ...................................................

    types_getDowntimes = [six.string_types, six.string_types, six.string_types]

    @classmethod
    def export_getDowntimes(cls, element, elementType, name):

        if elementType == "StorageElement":
            res = getSEHosts(name)
            if not res["OK"]:
                return res
            names = res["Value"]
        else:
            names = name

        return cls.rmClient.selectDowntimeCache(
            element=element, name=names, meta={"columns": ["StartDate", "EndDate", "Link", "Description", "Severity"]}
        )

    types_getCachedDowntimes = [
        (six.string_types, type(None), list),
        (six.string_types, type(None), list),
        (six.string_types, type(None), list),
        (six.string_types, type(None), list),
    ]

    def export_getCachedDowntimes(self, element, elementType, name, severity):

        if elementType == "StorageElement":
            res = getSEHosts(name)
            if not res["OK"]:
                return res
            names = res["Value"]
        else:
            names = name

        columns = ["Element", "Name", "StartDate", "EndDate", "Severity", "Description", "Link"]

        res = self.rmClient.selectDowntimeCache(
            element=element, name=names, severity=severity, meta={"columns": columns}
        )
        if not res["OK"]:
            self.log.error("Error selecting downtime cache", res["Message"])
            return res

        result = S_OK(res["Value"])
        result["Columns"] = columns

        return result

    types_setStatus = [six.string_types] * 7

    def export_setStatus(self, element, name, statusType, status, elementType, username, lastCheckTime):

        if not lastCheckTime:
            lastCheckTime = None
        else:
            lastCheckTime = datetime.strptime(lastCheckTime, "%Y-%m-%d %H:%M:%S")

        credentials = self.getRemoteCredentials()
        self.log.info(credentials)

        elementInDB = self.rsClient.selectStatusElement(
            element,
            "Status",
            name=name,
            statusType=statusType,  # status = status,
            elementType=elementType,
            lastCheckTime=lastCheckTime,
        )
        if not elementInDB["OK"]:
            self.log.error("Error selecting status elements", elementInDB["Message"])
            return elementInDB
        elif not elementInDB["Value"]:
            return S_ERROR("Your selection has been modified. Please refresh.")

        reason = "Status %s forced by %s ( web )" % (status, username)
        tokenExpiration = datetime.utcnow() + timedelta(days=1)

        newStatus = self.rsClient.addOrModifyStatusElement(
            element,
            "Status",
            name=name,
            statusType=statusType,
            status=status,
            elementType=elementType,
            reason=reason,
            tokenOwner=username,
            tokenExpiration=tokenExpiration,
        )
        if not newStatus["OK"]:
            self.log.error("Error setting status", newStatus["Message"])
            return newStatus

        return S_OK(reason)



""" DIRAC Basic MySQL Class
    It provides access to the basic MySQL methods in a multithread-safe mode
    keeping used connections in a python Queue for further reuse.

    These are the coded methods:


    __init__( host, user, passwd, name, [maxConnsInQueue=10] )

    Initializes the Queue and tries to connect to the DB server,
    using the _connect method.
    "maxConnsInQueue" defines the size of the Queue of open connections
    that are kept for reuse. It also defined the maximum number of open
    connections available from the object.
    maxConnsInQueue = 0 means unlimited and it is not supported.


    _except( methodName, exception, errorMessage )

    Helper method for exceptions: the "methodName" and the "errorMessage"
    are printed with ERROR level, then the "exception" is printed (with
    full description if it is a MySQL Exception) and S_ERROR is returned
    with the errorMessage and the exception.


    _connect()

    Attempts connection to DB and sets the _connected flag to True upon success.
    Returns S_OK or S_ERROR.


    _query( cmd, [conn] )

    Executes SQL command "cmd".
    Gets a connection from the Queue (or open a new one if none is available),
    the used connection is  back into the Queue.
    If a connection to the the DB is passed as second argument this connection
    is used and is not  in the Queue.
    Returns S_OK with fetchall() out in Value or S_ERROR upon failure.


    _update( cmd, [conn] )

    Executes SQL command "cmd" and issue a commit
    Gets a connection from the Queue (or open a new one if none is available),
    the used connection is  back into the Queue.
    If a connection to the the DB is passed as second argument this connection
    is used and is not  in the Queue
    Returns S_OK with number of updated registers in Value or S_ERROR upon failure.


    _createTables( tableDict )

    Create a new Table in the DB


    _getConnection()

    Gets a connection from the Queue (or open a new one if none is available)
    Returns S_OK with connection in Value or S_ERROR
    the calling method is responsible for closing this connection once it is no
    longer needed.




    Some high level methods have been added to avoid the need to write SQL
    statement in most common cases. They should be used instead of low level
    _insert, _update methods when ever possible.

    buildCondition( self, condDict = None, older = None, newer = None,
                      timeStamp = None, orderAttribute = None, limit = False,
                      greater = None, smaller = None ):

      Build SQL condition statement from provided condDict and other extra check on
      a specified time stamp.
      The conditions dictionary specifies for each attribute one or a List of possible
      values
      greater and smaller are dictionaries in which the keys are the names of the fields,
      that are requested to be >= or < than the corresponding value.
      For compatibility with current usage it uses Exceptions to exit in case of
      invalid arguments


    insertFields( self, tableName, inFields = None, inValues = None, conn = None, inDict = None ):

      Insert a new row in "tableName" assigning the values "inValues" to the
      fields "inFields".
      Alternatively inDict can be used
      String type values will be appropriately escaped.


    updateFields( self, tableName, updateFields = None, updateValues = None,
                  condDict = None,
                  limit = False, conn = None,
                  updateDict = None,
                  older = None, newer = None,
                  timeStamp = None, orderAttribute = None ):

      Update "updateFields" from "tableName" with "updateValues".
      updateDict alternative way to provide the updateFields and updateValues
      N records can match the condition
      return S_OK( number of updated rows )
      if limit is not False, the given limit is set
      String type values will be appropriately escaped.


    deleteEntries( self, tableName,
                   condDict = None,
                   limit = False, conn = None,
                   older = None, newer = None,
                   timeStamp = None, orderAttribute = None ):

      Delete rows from "tableName" with
      N records can match the condition
      if limit is not False, the given limit is set
      String type values will be appropriately escaped, they can be single values or lists of values.


    getFields( self, tableName, outFields = None,
               condDict = None,
               limit = False, conn = None,
               older = None, newer = None,
               timeStamp = None, orderAttribute = None ):

      Select "outFields" from "tableName" with condDict
      N records can match the condition
      return S_OK( tuple(Field,Value) )
      if limit is not False, the given limit is set
      String type values will be appropriately escaped, they can be single values or lists of values.

      for compatibility with other methods condDict keyed argument is added


    getCounters( self, table, attrList, condDict = None, older = None,
                 newer = None, timeStamp = None, connection = False ):

      Count the number of records on each distinct combination of AttrList, selected
      with condition defined by condDict and time stamps


    getDistinctAttributeValues( self, table, attribute, condDict = None, older = None,
                                newer = None, timeStamp = None, connection = False ):

      Get distinct values of a table attribute under specified conditions


"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import six
import collections
import time
import threading
import MySQLdb

from DIRAC import gLogger
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.Time import fromString
from DIRAC.Core.Utilities import DErrno

gInstancesCount = 0

__RCSID__ = "$Id$"

MAXCONNECTRETRY = 10
RETRY_SLEEP_DURATION = 5


def _checkFields(inFields, inValues):
    """
    Helper to check match between inFields and inValues lengths
    """

    if inFields is None and inValues is None:
        return S_OK()

    try:
        assert len(inFields) == len(inValues)
    except AssertionError:
        return S_ERROR(DErrno.EMYSQL, "Mismatch between inFields and inValues.")

    return S_OK()


def _quotedList(fieldList=None):
    """
    Quote a list of MySQL Field Names with "`"
    Return a comma separated list of quoted Field Names

    To be use for Table and Field Names
    """
    if fieldList is None:
        return None
    quotedFields = []
    try:
        for field in fieldList:
            quotedFields.append("`%s`" % field.replace("`", ""))
    except Exception:
        return None
    if not quotedFields:
        return None

    return ", ".join(quotedFields)


class ConnectionPool(object):
    """
    Management of connections per thread
    """

    def __init__(self, host, user, passwd, port=3306, graceTime=600):
        self.__host = host
        self.__user = user
        self.__passwd = passwd
        self.__port = port
        self.__graceTime = graceTime
        self.__spares = collections.deque()
        self.__maxSpares = 10
        self.__lastClean = 0
        self.__assigned = {}

    @property
    def __thid(self):
        return threading.current_thread()

    def __newConn(self):
        conn = MySQLdb.connect(host=self.__host, port=self.__port, user=self.__user, passwd=self.__passwd)

        self.__execute(conn, "SET AUTOCOMMIT=1")
        return conn

    def __execute(self, conn, cmd):
        cursor = conn.cursor()
        res = cursor.execute(cmd)
        conn.commit()
        cursor.close()
        return res

    def get(self, dbName, retries=10):
        retries = max(0, min(MAXCONNECTRETRY, retries))
        self.clean()
        return self.__getWithRetry(dbName, retries, retries)

    def __getWithRetry(self, dbName, totalRetries, retriesLeft):
        sleepTime = RETRY_SLEEP_DURATION * (totalRetries - retriesLeft)
        if sleepTime > 0:
            time.sleep(sleepTime)
        try:
            conn, lastName, thid = self.__innerGet()
        except MySQLdb.MySQLError as excp:
            if retriesLeft > 0:
                return self.__getWithRetry(dbName, totalRetries, retriesLeft - 1)
            return S_ERROR(DErrno.EMYSQL, "Could not connect: %s" % excp)

        if not self.__ping(conn):
            try:
                self.__assigned.pop(thid)
            except KeyError:
                pass
            if retriesLeft > 0:
                return self.__getWithRetry(dbName, totalRetries, retriesLeft)
            return S_ERROR(DErrno.EMYSQL, "Could not connect")

        if lastName != dbName:
            try:
                conn.select_db(dbName)
            except MySQLdb.MySQLError as excp:
                if retriesLeft > 0:
                    return self.__getWithRetry(dbName, totalRetries, retriesLeft - 1)
                return S_ERROR(DErrno.EMYSQL, "Could not select db %s: %s" % (dbName, excp))
            try:
                self.__assigned[thid][1] = dbName
            except KeyError:
                if retriesLeft > 0:
                    return self.__getWithRetry(dbName, totalRetries, retriesLeft - 1)
                return S_ERROR(DErrno.EMYSQL, "Could not connect")
        return S_OK(conn)

    def __ping(self, conn):
        try:
            conn.ping(True)
            return True
        except Exception:
            return False

    def __innerGet(self):
        thid = self.__thid
        now = time.time()
        if thid in self.__assigned:
            data = self.__assigned[thid]
            conn = data[0]
            data[2] = now
            return data[0], data[1], thid
        # Not cached
        try:
            conn, dbName = self.__spares.pop()
        except IndexError:
            conn = self.__newConn()
            dbName = ""

        self.__assigned[thid] = [conn, dbName, now]
        return conn, dbName, thid

    def __pop(self, thid):
        try:
            data = self.__assigned.pop(thid)
            if len(self.__spares) < self.__maxSpares:
                self.__spares.append((data[0], data[1]))
            else:
                try:
                    data[0].close()
                except MySQLdb.ProgrammingError as exc:
                    gLogger.warn("ProgrammingError exception while closing MySQL connection: %s" % exc)
                except Exception as exc:
                    gLogger.warn("Exception while closing MySQL connection: %s" % exc)
        except KeyError:
            pass

    def clean(self, now=False):
        if not now:
            now = time.time()
        self.__lastClean = now
        for thid in list(self.__assigned):
            if not thid.is_alive():
                self.__pop(thid)
            try:
                data = self.__assigned[thid]
            except KeyError:
                continue
            if now - data[2] > self.__graceTime:
                self.__pop(thid)

    def transactionStart(self, dbName):
        result = self.get(dbName)
        if not result["OK"]:
            return result
        conn = result["Value"]
        try:
            return S_OK(self.__execute(conn, "START TRANSACTION WITH CONSISTENT SNAPSHOT"))
        except MySQLdb.MySQLError as excp:
            return S_ERROR(DErrno.EMYSQL, "Could not begin transaction: %s" % excp)

    def transactionCommit(self, dbName):
        result = self.get(dbName)
        if not result["OK"]:
            return result
        conn = result["Value"]
        try:
            result = self.__execute(conn, "COMMIT")
            return S_OK(result)
        except MySQLdb.MySQLError as excp:
            return S_ERROR(DErrno.EMYSQL, "Could not commit transaction: %s" % excp)

    def transactionRollback(self, dbName):
        result = self.get(dbName)
        if not result["OK"]:
            return result
        conn = result["Value"]
        try:
            result = self.__execute(conn, "ROLLBACK")
            return S_OK(result)
        except MySQLdb.MySQLError as excp:
            return S_ERROR(DErrno.EMYSQL, "Could not rollback transaction: %s" % excp)


class MySQL(object):
    """
    Basic multithreaded DIRAC MySQL Client Class
    """

    __initialized = False

    __connectionPools = {}

    def __init__(self, hostName="localhost", userName="dirac", passwd="dirac", dbName="", port=3306, debug=False):
        """
        set MySQL connection parameters and try to connect

        :param debug: unused
        """
        global gInstancesCount
        gInstancesCount += 1

        self._connected = False

        if "log" not in dir(self):
            self.log = gLogger.getSubLogger("MySQL")
        self.logger = self.log

        # let the derived class decide what to do with if is not 1
        self._threadsafe = MySQLdb.threadsafety
        # self.log.debug('thread_safe = %s' % self._threadsafe)

        self.__hostName = str(hostName)
        self.__userName = str(userName)
        self.__passwd = str(passwd)
        self.__dbName = str(dbName)
        self.__port = port
        cKey = (self.__hostName, self.__userName, self.__passwd, self.__port)
        if cKey not in MySQL.__connectionPools:
            MySQL.__connectionPools[cKey] = ConnectionPool(*cKey)
        self.__connectionPool = MySQL.__connectionPools[cKey]

        self.__initialized = True
        result = self._connect()
        if not result["OK"]:
            gLogger.error("Cannot connect to the DB", " %s" % result["Message"])

    def __del__(self):
        global gInstancesCount
        try:
            gInstancesCount -= 1
        except Exception:
            pass

    def _except(self, methodName, x, err, cmd=""):
        """
        print MySQL error or exception
        return S_ERROR with Exception
        """

        try:
            raise x
        except MySQLdb.Error as e:
            self.log.error("%s (%s): %s" % (methodName, self._safeCmd(cmd), err), "%d: %s" % (e.args[0], e.args[1]))
            return S_ERROR(DErrno.EMYSQL, "%s: ( %d: %s )" % (err, e.args[0], e.args[1]))
        except Exception as e:
            self.log.error("%s (%s): %s" % (methodName, self._safeCmd(cmd), err), repr(e))
            return S_ERROR(DErrno.EMYSQL, "%s: (%s)" % (err, repr(e)))

    def __isDateTime(self, dateString):

        if dateString == "UTC_TIMESTAMP()":
            return True
        try:
            dtime = dateString.replace('"', "").replace("'", "")
            dtime = fromString(dtime)
            if dtime is None:
                return False
            return True
        except Exception:
            return False

    def __escapeString(self, myString, connection=None):
        """
        To be used for escaping any MySQL string before passing it to the DB
        this should prevent passing non-MySQL accepted characters to the DB
        It also includes quotation marks " around the given string
        """
        if connection is None:
            retDict = self._getConnection()
            if not retDict["OK"]:
                return retDict
            connection = retDict["Value"]

        if six.PY3 and isinstance(myString, bytes):
            myString = myString.decode()
        try:
            myString = str(myString)
        except ValueError:
            return S_ERROR(DErrno.EMYSQL, "Cannot escape value!")

        timeUnits = ["MICROSECOND", "SECOND", "MINUTE", "HOUR", "DAY", "WEEK", "MONTH", "QUARTER", "YEAR"]

        try:
            # Check datetime functions first
            if myString.strip() == "UTC_TIMESTAMP()":
                return S_OK(myString)

            for func in ["TIMESTAMPDIFF", "TIMESTAMPADD"]:
                if myString.strip().startswith("%s(" % func) and myString.strip().endswith(")"):
                    args = myString.strip()[:-1].replace("%s(" % func, "").strip().split(",")
                    arg1, arg2, arg3 = [x.strip() for x in args]
                    if arg1 in timeUnits:
                        if self.__isDateTime(arg2) or arg2.isalnum():
                            if self.__isDateTime(arg3) or arg3.isalnum():
                                return S_OK(myString)
                    # self.log.debug('__escape_string: Could not escape string', '"%s"' % myString)
                    return S_ERROR(DErrno.EMYSQL, "__escape_string: Could not escape string")

            if six.PY3:
                escape_string = connection.escape_string(myString.encode()).decode()
            else:
                escape_string = connection.escape_string(myString)
            # self.log.debug('__escape_string: returns', '"%s"' % escape_string)
            return S_OK('"%s"' % escape_string)
        except Exception as x:
            return self._except("__escape_string", x, "Could not escape string", myString)

    def __checkTable(self, tableName, force=False):
        """Check if a table exists by issuing 'SHOW TABLES'

        :param str tableName: table name in check
        :param bool force: force or not the re-creation (would drop the previous one)
        :returns: S_OK/S_ERROR
        """

        table = _quotedList([tableName])
        if not table:
            return S_ERROR(DErrno.EMYSQL, "Invalid tableName argument")

        cmd = "SHOW TABLES"
        retDict = self._query(cmd)
        if not retDict["OK"]:
            return retDict
        if (tableName,) in retDict["Value"]:
            if not force:
                # the requested exist and table creation is not force, return with error
                return S_ERROR(DErrno.EMYSQL, "The requested table already exist")
            else:
                cmd = "DROP TABLE %s" % table
                retDict = self._update(cmd)
                if not retDict["OK"]:
                    return retDict

        return S_OK()

    def _escapeString(self, myString, conn=None):
        """
        Wrapper around the internal method __escapeString
        """
        # self.log.debug('_escapeString:', '"%s"' % str(myString))

        return self.__escapeString(myString)

    def _escapeValues(self, inValues=None):
        """
        Escapes all strings in the list of values provided
        """
        # self.log.debug('_escapeValues:', inValues)
        retDict = self._getConnection()
        if not retDict["OK"]:
            return retDict
        connection = retDict["Value"]

        inEscapeValues = []

        if not inValues:
            return S_OK(inEscapeValues)

        for value in inValues:
            if isinstance(value, six.string_types):
                retDict = self.__escapeString(value, connection=connection)
                if not retDict["OK"]:
                    return retDict
                inEscapeValues.append(retDict["Value"])
            elif isinstance(value, (tuple, list)):
                tupleValues = []
                for val in value:
                    retDict = self.__escapeString(val, connection=connection)
                    if not retDict["OK"]:
                        return retDict
                    tupleValues.append(retDict["Value"])
                inEscapeValues.append("(" + ", ".join(tupleValues) + ")")
            elif isinstance(value, bool):
                inEscapeValues = [str(value)]
            else:
                if six.PY3 and isinstance(value, bytes):
                    value = value.decode()
                retDict = self.__escapeString(str(value), connection=connection)
                if not retDict["OK"]:
                    return retDict
                inEscapeValues.append(retDict["Value"])
        return S_OK(inEscapeValues)

    def _safeCmd(self, command):
        """Just replaces password, if visible, with *********"""
        return command.replace(self.__passwd, "**********")

    def _connect(self):
        """
        open connection to MySQL DB and put Connection into Queue
        set connected flag to True and return S_OK
        return S_ERROR upon failure
        """
        if not self.__initialized:
            error = "DB not properly initialized"
            gLogger.error(error)
            return S_ERROR(DErrno.EMYSQL, error)

        # self.log.debug('_connect:', self._connected)
        if self._connected:
            return S_OK()

        # Test the connection to the DB
        retDict = self._getConnection()
        if not retDict["OK"]:
            return retDict
        self._connected = True
        return S_OK()

    def _query(self, cmd, conn=None, debug=False):
        """
        execute MySQL query command

        :param debug: unused

        return S_OK structure with fetchall result as tuple
        it returns an empty tuple if no matching rows are found
        return S_ERROR upon error
        """

        self.log.debug("_query: %s" % self._safeCmd(cmd))

        retDict = self._getConnection()
        if not retDict["OK"]:
            return retDict
        connection = retDict["Value"]

        try:
            cursor = connection.cursor()
            if cursor.execute(cmd):
                res = cursor.fetchall()
            else:
                res = ()

            # Log the result limiting it to just 10 records
            # if len(res) <= 10:
            #  self.log.debug('_query: returns', res)
            # else:
            #  self.log.debug('_query: Total %d records returned' % len(res))
            #  self.log.debug('_query: %s ...' % str(res[:10]))

            retDict = S_OK(res)
        except Exception as x:
            # self.log.debug('_query: %s' % self._safeCmd(cmd))
            retDict = self._except("_query", x, "Execution failed.", cmd)

        try:
            cursor.close()
        except Exception:
            pass

        return retDict

    def _update(self, cmd, conn=None, debug=False):
        """execute MySQL update command

        :param debug: unused

        return S_OK with number of updated registers upon success
        return S_ERROR upon error
        """

        self.log.debug("_update: %s" % self._safeCmd(cmd))

        retDict = self._getConnection()
        if not retDict["OK"]:
            return retDict
        connection = retDict["Value"]

        try:
            cursor = connection.cursor()
            res = cursor.execute(cmd)
            retDict = S_OK(res)
            if cursor.lastrowid:
                retDict["lastRowId"] = cursor.lastrowid
        except Exception as x:
            retDict = self._except("_update", x, "Execution failed.", cmd)

        try:
            cursor.close()
        except Exception:
            pass

        return retDict

    def _transaction(self, cmdList, conn=None):
        """dummy transaction support

        :param self: self reference
        :param list cmdList: list of queries to be executed within the transaction
        :param MySQLDB.Connection conn: connection

        :return: S_OK( [ ( cmd1, ret1 ), ... ] ) or S_ERROR
        """
        if not isinstance(cmdList, list):
            return S_ERROR(DErrno.EMYSQL, "_transaction: wrong type (%s) for cmdList" % type(cmdList))

        # # get connection
        connection = conn
        if not connection:
            retDict = self._getConnection()
            if not retDict["OK"]:
                return retDict
            connection = retDict["Value"]

        # # list with cmds and their results
        cmdRet = []
        try:
            cursor = connection.cursor()
            for cmd in cmdList:
                cmdRet.append((cmd, cursor.execute(cmd)))
            connection.commit()
        except Exception as error:
            self.logger.exception(error)
            # # rollback, put back connection to the pool
            connection.rollback()
            return S_ERROR(DErrno.EMYSQL, error)
        # # close cursor, put back connection to the pool
        cursor.close()
        return S_OK(cmdRet)

    def _createViews(self, viewsDict, force=False):
        """create view based on query

        :param dict viewDict: { 'ViewName': "Fields" : { "`a`": `tblA.a`, "`sumB`" : "SUM(`tblB.b`)" }
                                            "SelectFrom" : "tblA join tblB on tblA.id = tblB.id",
                                            "Clauses" : [ "`tblA.a` > 10", "`tblB.Status` = 'foo'" ] ## WILL USE AND CLAUSE
                                            "GroupBy": [ "`a`" ],
                                            "OrderBy": [ "`b` DESC" ] }
        """
        if force:
            # gLogger.debug(viewsDict)

            for viewName, viewDict in viewsDict.items():

                viewQuery = ["CREATE OR REPLACE VIEW `%s`.`%s` AS" % (self.__dbName, viewName)]

                columns = ",".join(
                    ["%s AS %s" % (colDef, colName) for colName, colDef in viewDict.get("Fields", {}).items()]
                )
                tables = viewDict.get("SelectFrom", "")
                if columns and tables:
                    viewQuery.append("SELECT %s FROM %s" % (columns, tables))

                where = " AND ".join(viewDict.get("Clauses", []))
                if where:
                    viewQuery.append("WHERE %s" % where)

                groupBy = ",".join(viewDict.get("GroupBy", []))
                if groupBy:
                    viewQuery.append("GROUP BY %s" % groupBy)

                orderBy = ",".join(viewDict.get("OrderBy", []))
                if orderBy:
                    viewQuery.append("ORDER BY %s" % orderBy)

                viewQuery.append(";")
                viewQuery = " ".join(viewQuery)
                # self.log.debug("`%s` VIEW QUERY IS: %s" % (viewName, viewQuery))
                createView = self._query(viewQuery)
                if not createView["OK"]:
                    self.log.error("Can not create view", createView["Message"])
                    return createView
        return S_OK()

    def _createTables(self, tableDict, force=False):
        """
        tableDict:
          tableName: { 'Fields' : { 'Field': 'Description' },
                       'ForeignKeys': {'Field': 'Table.key' },
                       'PrimaryKey': 'Id',
                       'Indexes': { 'Index': [] },
                       'UniqueIndexes': { 'Index': [] },
                       'Engine': 'InnoDB' }
          only 'Fields' is a mandatory key.

        Creates a new Table for each key in tableDict, "tableName" in the DB with
        the provided description.
        It allows to create:
          - flat tables if no "ForeignKeys" key defined.
          - tables with foreign keys to auxiliary tables holding the values
          of some of the fields
        Arguments:
          tableDict: dictionary of dictionary with description of tables to be created.
          Only "Fields" is a mandatory key in the table description.
            "Fields": Dictionary with Field names and description of the fields
            "ForeignKeys": Dictionary with Field names and name of auxiliary tables.
              The auxiliary tables must be defined in tableDict.
            "PrimaryKey": Name of PRIMARY KEY for the table (if exist).
            "Indexes": Dictionary with definition of indexes, the value for each
              index is the list of fields to be indexed.
            "UniqueIndexes": Dictionary with definition of indexes, the value for each
              index is the list of fields to be indexed. This indexes will declared
              unique.
            "Engine": use the given DB engine, InnoDB is the default if not present.
            "Charset": use the given character set. Default is latin1
          force:
            if True, requested tables are DROP if they exist.
            if False, returned with S_ERROR if table exist.

        """

        # First check consistency of request
        if not isinstance(tableDict, dict):
            return S_ERROR(DErrno.EMYSQL, "Argument is not a dictionary: %s( %s )" % (type(tableDict), tableDict))

        tableList = list(tableDict)
        if len(tableList) == 0:
            return S_OK(0)
        for table in tableList:
            thisTable = tableDict[table]
            # Check if Table is properly described with a dictionary
            if not isinstance(thisTable, dict):
                return S_ERROR(
                    DErrno.EMYSQL, "Table description is not a dictionary: %s( %s )" % (type(thisTable), thisTable)
                )
            if "Fields" not in thisTable:
                return S_ERROR(DErrno.EMYSQL, "Missing `Fields` key in `%s` table dictionary" % table)

        tableCreationList = [[]]

        auxiliaryTableList = []

        i = 0
        extracted = True
        while tableList and extracted:
            # iterate extracting tables from list if they only depend on
            # already extracted tables.
            extracted = False
            auxiliaryTableList += tableCreationList[i]
            i += 1
            tableCreationList.append([])
            for table in list(tableList):
                toBeExtracted = True
                thisTable = tableDict[table]
                if "ForeignKeys" in thisTable:
                    thisKeys = thisTable["ForeignKeys"]
                    for key, auxTable in thisKeys.items():
                        forTable = auxTable.split(".")[0]
                        forKey = key
                        if forTable != auxTable:
                            forKey = auxTable.split(".")[1]
                        if forTable not in auxiliaryTableList:
                            toBeExtracted = False
                            break
                        if key not in thisTable["Fields"]:
                            return S_ERROR(
                                DErrno.EMYSQL,
                                "ForeignKey `%s` -> `%s` not defined in Primary table `%s`." % (key, forKey, table),
                            )
                        if forKey not in tableDict[forTable]["Fields"]:
                            return S_ERROR(
                                DErrno.EMYSQL,
                                "ForeignKey `%s` -> `%s` not defined in Auxiliary table `%s`."
                                % (key, forKey, forTable),
                            )

                if toBeExtracted:
                    # self.log.debug('Table %s ready to be created' % table)
                    extracted = True
                    tableList.remove(table)
                    tableCreationList[i].append(table)

        if tableList:
            return S_ERROR(DErrno.EMYSQL, "Recursive Foreign Keys in %s" % ", ".join(tableList))

        for tableList in tableCreationList:
            for table in tableList:
                # Check if Table exist
                retDict = self.__checkTable(table, force=force)
                if not retDict["OK"]:
                    return retDict

                thisTable = tableDict[table]
                cmdList = []
                for field in thisTable["Fields"].keys():
                    cmdList.append("`%s` %s" % (field, thisTable["Fields"][field]))

                if "PrimaryKey" in thisTable:
                    if isinstance(thisTable["PrimaryKey"], six.string_types):
                        cmdList.append("PRIMARY KEY ( `%s` )" % thisTable["PrimaryKey"])
                    else:
                        cmdList.append(
                            "PRIMARY KEY ( %s )" % ", ".join(["`%s`" % str(f) for f in thisTable["PrimaryKey"]])
                        )

                if "Indexes" in thisTable:
                    indexDict = thisTable["Indexes"]
                    for index in indexDict:
                        indexedFields = "`, `".join(indexDict[index])
                        cmdList.append("INDEX `%s` ( `%s` )" % (index, indexedFields))

                if "UniqueIndexes" in thisTable:
                    indexDict = thisTable["UniqueIndexes"]
                    for index in indexDict:
                        indexedFields = "`, `".join(indexDict[index])
                        cmdList.append("UNIQUE INDEX `%s` ( `%s` )" % (index, indexedFields))
                if "ForeignKeys" in thisTable:
                    thisKeys = thisTable["ForeignKeys"]
                    for key, auxTable in thisKeys.items():

                        forTable = auxTable.split(".")[0]
                        forKey = key
                        if forTable != auxTable:
                            forKey = auxTable.split(".")[1]

                        # cmdList.append( '`%s` %s' % ( forTable, tableDict[forTable]['Fields'][forKey] )
                        cmdList.append(
                            "FOREIGN KEY ( `%s` ) REFERENCES `%s` ( `%s` )"
                            " ON DELETE RESTRICT" % (key, forTable, forKey)
                        )

                engine = thisTable.get("Engine", "InnoDB")
                charset = thisTable.get("Charset", "latin1")

                cmd = "CREATE TABLE `%s` (\n%s\n) ENGINE=%s DEFAULT CHARSET=%s" % (
                    table,
                    ",\n".join(cmdList),
                    engine,
                    charset,
                )
                retDict = self._transaction([cmd])
                if not retDict["OK"]:
                    return retDict
                # self.log.debug('Table %s created' % table)

        return S_OK()

    def _to_value(self, param):
        """
        Convert to string
        """
        return str(param[0])

    def _getConnection(self, retries=MAXCONNECTRETRY):
        """Return  a new connection to the DB,

        Try the Queue, if it is empty add a newConnection to the Queue and retry
        it will retry MAXCONNECTRETRY to open a new connection and will return
        an error if it fails.

        :param int retries: Number of time it will retry to open a connection
        """
        # self.log.debug('_getConnection:')

        if not self.__initialized:
            error = "DB not properly initialized"
            gLogger.error(error)
            return S_ERROR(DErrno.EMYSQL, error)

        return self.__connectionPool.get(self.__dbName, retries)

    ########################################################################################
    #
    #  Transaction functions
    #
    ########################################################################################

    def transactionStart(self):
        return self.__connectionPool.transactionStart(self.__dbName)

    def transactionCommit(self):
        return self.__connectionPool.transactionCommit(self.__dbName)

    def transactionRollback(self):
        return self.__connectionPool.transactionRollback(self.__dbName)

    ########################################################################################
    #
    #  Utility functions
    #
    ########################################################################################

    def countEntries(
        self, table, condDict, older=None, newer=None, timeStamp=None, connection=False, greater=None, smaller=None
    ):
        """
        Count the number of entries wit the given conditions
        """
        table = _quotedList([table])
        if not table:
            error = "Invalid table argument"
            # self.log.debug('countEntries:', error)
            return S_ERROR(DErrno.EMYSQL, error)

        try:
            cond = self.buildCondition(
                condDict=condDict, older=older, newer=newer, timeStamp=timeStamp, greater=greater, smaller=smaller
            )
        except Exception as x:
            return S_ERROR(DErrno.EMYSQL, x)

        cmd = "SELECT COUNT(*) FROM %s %s" % (table, cond)
        res = self._query(cmd, connection)
        if not res["OK"]:
            return res

        return S_OK(res["Value"][0][0])

    ########################################################################################
    def getCounters(
        self,
        table,
        attrList,
        condDict,
        older=None,
        newer=None,
        timeStamp=None,
        connection=False,
        greater=None,
        smaller=None,
    ):
        """
        Count the number of records on each distinct combination of AttrList, selected
        with condition defined by condDict and time stamps
        """
        table = _quotedList([table])
        if not table:
            error = "Invalid table argument"
            # self.log.debug('getCounters:', error)
            return S_ERROR(DErrno.EMYSQL, error)

        attrNames = _quotedList(attrList)
        if attrNames is None:
            error = "Invalid updateFields argument"
            # self.log.debug('getCounters:', error)
            return S_ERROR(DErrno.EMYSQL, error)

        try:
            cond = self.buildCondition(
                condDict=condDict, older=older, newer=newer, timeStamp=timeStamp, greater=greater, smaller=smaller
            )
        except Exception as x:
            return S_ERROR(DErrno.EMYSQL, x)

        cmd = "SELECT %s, COUNT(*) FROM %s %s GROUP BY %s ORDER BY %s" % (attrNames, table, cond, attrNames, attrNames)
        res = self._query(cmd, connection)
        if not res["OK"]:
            return res

        resultList = []
        for raw in res["Value"]:
            attrDict = {}
            for ind, attr in enumerate(attrList):
                attrDict[attr] = raw[ind]
            item = (attrDict, raw[len(attrList)])
            resultList.append(item)
        return S_OK(resultList)

    #########################################################################################
    def getDistinctAttributeValues(
        self,
        table,
        attribute,
        condDict=None,
        older=None,
        newer=None,
        timeStamp=None,
        connection=False,
        greater=None,
        smaller=None,
    ):
        """
        Get distinct values of a table attribute under specified conditions
        """
        table = _quotedList([table])
        if not table:
            error = "Invalid table argument"
            # self.log.debug('getDistinctAttributeValues:', error)
            return S_ERROR(DErrno.EMYSQL, error)

        attributeName = _quotedList([attribute])
        if not attributeName:
            error = "Invalid attribute argument"
            # self.log.debug('getDistinctAttributeValues:', error)
            return S_ERROR(DErrno.EMYSQL, error)

        try:
            cond = self.buildCondition(
                condDict=condDict, older=older, newer=newer, timeStamp=timeStamp, greater=greater, smaller=smaller
            )
        except Exception as exc:
            return S_ERROR(DErrno.EMYSQL, exc)

        cmd = "SELECT DISTINCT( %s ) FROM %s %s ORDER BY %s" % (attributeName, table, cond, attributeName)
        res = self._query(cmd, connection)
        if not res["OK"]:
            return res
        attr_list = [x[0] for x in res["Value"]]
        return S_OK(attr_list)

    #############################################################################
    def buildCondition(
        self,
        condDict=None,
        older=None,
        newer=None,
        timeStamp=None,
        orderAttribute=None,
        limit=False,
        greater=None,
        smaller=None,
        offset=None,
        useLikeQuery=False,
    ):
        """Build SQL condition statement from provided condDict and other extra check on
        a specified time stamp.
        The conditions dictionary specifies for each attribute one or a List of possible
        values
        greater and smaller are dictionaries in which the keys are the names of the fields,
        that are requested to be >= or < than the corresponding value.
        For compatibility with current usage it uses Exceptions to exit in case of
        invalid arguments
        For performing LIKE queries use the parameter useLikeQuery=True
        """
        condition = ""
        conjunction = "WHERE"

        if condDict is not None:
            for aName, attrValue in condDict.items():
                if isinstance(aName, six.string_types):
                    attrName = _quotedList([aName])
                elif isinstance(aName, tuple):
                    attrName = "(" + _quotedList(list(aName)) + ")"
                if not attrName:
                    error = "Invalid condDict argument"
                    # self.log.debug('buildCondition:', error)
                    raise Exception(error)
                if isinstance(attrValue, list):
                    retDict = self._escapeValues(attrValue)
                    if not retDict["OK"]:
                        # self.log.debug('buildCondition:', retDict['Message'])
                        raise Exception(retDict["Message"])
                    else:
                        escapeInValues = retDict["Value"]
                        multiValue = ", ".join(escapeInValues)
                        condition = " %s %s %s IN ( %s )" % (condition, conjunction, attrName, multiValue)
                        conjunction = "AND"
                else:
                    retDict = self._escapeValues([attrValue])
                    if not retDict["OK"]:
                        # self.log.debug('buildCondition:', retDict['Message'])
                        raise Exception(retDict["Message"])
                    else:
                        escapeInValue = retDict["Value"][0]
                        if useLikeQuery:
                            condition = " %s %s %s LIKE %s" % (condition, conjunction, attrName, escapeInValue)
                        else:
                            if escapeInValue == "NULL":
                                condition = " %s %s %s IS NULL" % (condition, conjunction, attrName)
                            else:
                                condition = " %s %s %s = %s" % (condition, conjunction, attrName, escapeInValue)
                        conjunction = "AND"

        if timeStamp:
            timeStamp = _quotedList([timeStamp])
            if not timeStamp:
                error = "Invalid timeStamp argument"
                # self.log.debug('buildCondition:', error)
                raise Exception(error)
            if newer:
                retDict = self._escapeValues([newer])
                if not retDict["OK"]:
                    # self.log.debug('buildCondition:', retDict['Message'])
                    raise Exception(retDict["Message"])
                else:
                    escapeInValue = retDict["Value"][0]
                    condition = " %s %s %s >= %s" % (condition, conjunction, timeStamp, escapeInValue)
                    conjunction = "AND"
            if older:
                retDict = self._escapeValues([older])
                if not retDict["OK"]:
                    # self.log.debug('buildCondition:', retDict['Message'])
                    raise Exception(retDict["Message"])
                else:
                    escapeInValue = retDict["Value"][0]
                    condition = " %s %s %s < %s" % (condition, conjunction, timeStamp, escapeInValue)

        if isinstance(greater, dict):
            for attrName, attrValue in greater.items():
                attrName = _quotedList([attrName])
                if not attrName:
                    error = "Invalid greater argument"
                    # self.log.debug('buildCondition:', error)
                    raise Exception(error)

                retDict = self._escapeValues([attrValue])
                if not retDict["OK"]:
                    # self.log.debug('buildCondition:', retDict['Message'])
                    raise Exception(retDict["Message"])
                else:
                    escapeInValue = retDict["Value"][0]
                    condition = " %s %s %s >= %s" % (condition, conjunction, attrName, escapeInValue)
                    conjunction = "AND"

        if isinstance(smaller, dict):
            for attrName, attrValue in smaller.items():
                attrName = _quotedList([attrName])
                if not attrName:
                    error = "Invalid smaller argument"
                    # self.log.debug('buildCondition:', error)
                    raise Exception(error)

                retDict = self._escapeValues([attrValue])
                if not retDict["OK"]:
                    # self.log.debug('buildCondition:', retDict['Message'])
                    raise Exception(retDict["Message"])
                else:
                    escapeInValue = retDict["Value"][0]
                    condition = " %s %s %s < %s" % (condition, conjunction, attrName, escapeInValue)
                    conjunction = "AND"

        orderList = []
        orderAttrList = orderAttribute
        if not isinstance(orderAttrList, list):
            orderAttrList = [orderAttribute]
        for orderAttr in orderAttrList:
            if orderAttr is None:
                continue
            if not isinstance(orderAttr, six.string_types):
                error = "Invalid orderAttribute argument"
                # self.log.debug('buildCondition:', error)
                raise Exception(error)

            orderField = _quotedList(orderAttr.split(":")[:1])
            if not orderField:
                error = "Invalid orderAttribute argument"
                # self.log.debug('buildCondition:', error)
                raise Exception(error)

            if len(orderAttr.split(":")) == 2:
                orderType = orderAttr.split(":")[1].upper()
                if orderType in ["ASC", "DESC"]:
                    orderList.append("%s %s" % (orderField, orderType))
                else:
                    error = "Invalid orderAttribute argument"
                    # self.log.debug('buildCondition:', error)
                    raise Exception(error)
            else:
                orderList.append(orderAttr)

        if orderList:
            condition = "%s ORDER BY %s" % (condition, ", ".join(orderList))

        if limit:
            if offset:
                condition = "%s LIMIT %d OFFSET %d" % (condition, limit, offset)
            else:
                condition = "%s LIMIT %d" % (condition, limit)

        return condition

    #############################################################################
    def getFields(
        self,
        tableName,
        outFields=None,
        condDict=None,
        limit=False,
        conn=None,
        older=None,
        newer=None,
        timeStamp=None,
        orderAttribute=None,
        greater=None,
        smaller=None,
        useLikeQuery=False,
    ):
        """
        Select "outFields" from "tableName" with condDict
        N records can match the condition
        return S_OK(tuple(Field, Value))
        if outFields is None all fields in "tableName" are returned
        if limit is not False, the given limit is set
        inValues are properly escaped using the _escape_string method, they can be single values or lists of values.
        if useLikeQuery=True, then conDict can return matched rows if "%" is defined inside conDict.
        """
        table = _quotedList([tableName])
        if not table:
            error = "Invalid tableName argument"
            # self.log.debug('getFields:', error)
            return S_ERROR(DErrno.EMYSQL, error)

        quotedOutFields = "*"
        if outFields:
            quotedOutFields = _quotedList(outFields)
            if quotedOutFields is None:
                error = "Invalid outFields arguments"
                # self.log.debug('getFields:', error)
                return S_ERROR(DErrno.EMYSQL, error)

        # self.log.debug('getFields:', 'selecting fields %s from table %s.' % (quotedOutFields, table))

        if condDict is None:
            condDict = {}

        try:
            try:
                mylimit = limit[0]
                myoffset = limit[1]
            except TypeError:
                mylimit = limit
                myoffset = None
            condition = self.buildCondition(
                condDict=condDict,
                older=older,
                newer=newer,
                timeStamp=timeStamp,
                orderAttribute=orderAttribute,
                limit=mylimit,
                greater=greater,
                smaller=smaller,
                offset=myoffset,
                useLikeQuery=useLikeQuery,
            )
        except Exception as x:
            return S_ERROR(DErrno.EMYSQL, x)

        return self._query("SELECT %s FROM %s %s" % (quotedOutFields, table, condition), conn)

    #############################################################################
    def deleteEntries(
        self,
        tableName,
        condDict=None,
        limit=False,
        conn=None,
        older=None,
        newer=None,
        timeStamp=None,
        orderAttribute=None,
        greater=None,
        smaller=None,
    ):
        """
        Delete rows from "tableName" with
        N records can match the condition
        if limit is not False, the given limit is set
        String type values will be appropriately escaped, they can be single values or lists of values.
        """
        table = _quotedList([tableName])
        if not table:
            error = "Invalid tableName argument"
            # self.log.debug('deleteEntries:', error)
            return S_ERROR(DErrno.EMYSQL, error)

        # self.log.debug('deleteEntries:', 'deleting rows from table %s.' % table)

        try:
            condition = self.buildCondition(
                condDict=condDict,
                older=older,
                newer=newer,
                timeStamp=timeStamp,
                orderAttribute=orderAttribute,
                limit=limit,
                greater=greater,
                smaller=smaller,
            )
        except Exception as x:
            return S_ERROR(DErrno.EMYSQL, x)

        return self._update("DELETE FROM %s %s" % (table, condition), conn)

    #############################################################################
    def updateFields(
        self,
        tableName,
        updateFields=None,
        updateValues=None,
        condDict=None,
        limit=False,
        conn=None,
        updateDict=None,
        older=None,
        newer=None,
        timeStamp=None,
        orderAttribute=None,
        greater=None,
        smaller=None,
    ):
        """
        Update "updateFields" from "tableName" with "updateValues".
        updateDict alternative way to provide the updateFields and updateValues
        N records can match the condition
        return S_OK( number of updated rows )
        if limit is not False, the given limit is set
        String type values will be appropriately escaped.

        """
        if not updateFields and not updateDict:
            return S_OK(0)

        table = _quotedList([tableName])
        if not table:
            error = "Invalid tableName argument"
            # self.log.debug('updateFields:', error)
            return S_ERROR(DErrno.EMYSQL, error)

        retDict = _checkFields(updateFields, updateValues)
        if not retDict["OK"]:
            error = "Mismatch between updateFields and updateValues."
            # self.log.debug('updateFields:', error)
            return S_ERROR(DErrno.EMYSQL, error)

        if updateFields is None:
            updateFields = []
            updateValues = []

        if updateDict:
            if not isinstance(updateDict, dict):
                error = "updateDict must be a of Type DictType"
                # self.log.debug('updateFields:', error)
                return S_ERROR(DErrno.EMYSQL, error)
            try:
                updateFields += list(updateDict)
                updateValues += [updateDict[k] for k in updateFields]
            except TypeError:
                error = "updateFields and updateValues must be a list"
                # self.log.debug('updateFields:', error)
                return S_ERROR(DErrno.EMYSQL, error)

        updateValues = self._escapeValues(updateValues)
        if not updateValues["OK"]:
            # self.log.debug('updateFields:', updateValues['Message'])
            return updateValues
        updateValues = updateValues["Value"]

        # self.log.debug('updateFields:', 'updating fields %s from table %s.' % (', '.join(updateFields), table))

        try:
            condition = self.buildCondition(
                condDict=condDict,
                older=older,
                newer=newer,
                timeStamp=timeStamp,
                orderAttribute=orderAttribute,
                limit=limit,
                greater=greater,
                smaller=smaller,
            )
        except Exception as x:
            return S_ERROR(DErrno.EMYSQL, x)

        updateString = ",".join(
            ["%s = %s" % (_quotedList([updateFields[k]]), updateValues[k]) for k in range(len(updateFields))]
        )

        return self._update("UPDATE %s SET %s %s" % (table, updateString, condition), conn)

    #############################################################################
    def insertFields(self, tableName, inFields=None, inValues=None, conn=None, inDict=None):
        """
        Insert a new row in "tableName" assigning the values "inValues" to the
        fields "inFields".
        String type values will be appropriately escaped.
        """
        table = _quotedList([tableName])
        if not table:
            error = "Invalid tableName argument"
            # self.log.debug('insertFields:', error)
            return S_ERROR(DErrno.EMYSQL, error)

        retDict = _checkFields(inFields, inValues)
        if not retDict["OK"]:
            # self.log.debug('insertFields:', retDict['Message'])
            return retDict

        if inFields is None:
            inFields = []
            inValues = []

        if inDict:
            if not isinstance(inDict, dict):
                error = "inDict must be a of Type DictType"
                # self.log.debug('insertFields:', error)
                return S_ERROR(DErrno.EMYSQL, error)
            try:
                inFields += list(inDict)
                inValues += [inDict[k] for k in inFields]
            except TypeError:
                error = "inFields and inValues must be a list"
                # self.log.debug('insertFields:', error)
                return S_ERROR(DErrno.EMYSQL, error)

        inFieldString = _quotedList(inFields)
        if inFieldString is None:
            error = "Invalid inFields arguments"
            # self.log.debug('insertFields:', error)
            return S_ERROR(DErrno.EMYSQL, error)

        inFieldString = "(  %s )" % inFieldString

        retDict = self._escapeValues(inValues)
        if not retDict["OK"]:
            # self.log.debug('insertFields:', retDict['Message'])
            return retDict
        inValueString = ", ".join(retDict["Value"])
        inValueString = "(  %s )" % inValueString

        # self.log.debug('insertFields:', 'inserting %s into table %s'
        #               % (inFieldString, table))

        return self._update("INSERT INTO %s %s VALUES %s" % (table, inFieldString, inValueString), conn)

    def executeStoredProcedure(self, packageName, parameters, outputIds):
        conDict = self._getConnection()
        if not conDict["OK"]:
            return conDict

        connection = conDict["Value"]
        cursor = connection.cursor()
        try:
            cursor.callproc(packageName, parameters)
            row = []
            for oId in outputIds:
                resName = "@_%s_%s" % (packageName, oId)
                cursor.execute("SELECT %s" % resName)
                row.append(cursor.fetchone()[0])
            retDict = S_OK(row)
        except Exception as x:
            retDict = self._except("_query", x, "Execution failed.", packageName)
            connection.rollback()

        try:
            cursor.close()
        except Exception:
            pass
        return retDict

    # For the procedures that execute a select without storing the result
    def executeStoredProcedureWithCursor(self, packageName, parameters):
        conDict = self._getConnection()
        if not conDict["OK"]:
            return conDict

        connection = conDict["Value"]
        cursor = connection.cursor()
        try:
            #       execStr = "call %s(%s);" % ( packageName, ",".join( map( str, parameters ) ) )
            execStr = "call %s(%s);" % (
                packageName,
                ",".join(
                    ['"%s"' % param if isinstance(param, six.string_types) else str(param) for param in parameters]
                ),
            )
            cursor.execute(execStr)
            rows = cursor.fetchall()
            retDict = S_OK(rows)
        except Exception as x:
            retDict = self._except("_query", x, "Execution failed.", packageName)
            connection.rollback()
        try:
            cursor.close()
        except Exception:
            pass

        return retDict



#!/usr/bin/env python
"""
  Get VM instances available in the configured cloud sites
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

__RCSID__ = "$Id$"

from DIRAC.Core.Base import Script
from DIRAC import gLogger, exit as DIRACExit
from DIRAC.Core.Utilities.DIRACScript import DIRACScript

site = None
ce = None
image = None
voName = None


def setCE(args):
    global ce
    ce = args


def setSite(args):
    global Site
    Site = args


def setImage(args):
    global image
    image = args


def setVO(args):
    global vo
    vo = args


@DIRACScript()
def main():
    Script.setUsageMessage(
        "\n".join(
            [
                "Get VM nodes information",
                "Usage:",
                "%s [option]... [cfgfile]" % Script.scriptName,
                "Arguments:",
                " cfgfile: DIRAC Cfg with description of the configuration (optional)",
            ]
        )
    )
    Script.registerSwitch("S:", "Site=", "Site Name", setSite)
    Script.registerSwitch("C:", "CE=", "Cloud Endpoint Name ", setCE)
    Script.registerSwitch("I:", "Image=", "Image Name", setImage)
    Script.registerSwitch("v:", "vo=", "VO name", setVO)
    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getExtraCLICFGFiles()

    from DIRAC.WorkloadManagementSystem.Client.VMClient import VMClient
    from DIRAC.Core.Security.ProxyInfo import getVOfromProxyGroup
    from DIRAC.Core.Utilities.PrettyPrint import printTable

    siteList = None
    if site is not None:
        siteList = [s.strip() for s in site.split(",")]

    ceList = None
    if ce is not None:
        ceList = [c.strip() for c in ce.split(",")]

    voName = vo
    if voName is None:
        result = getVOfromProxyGroup()
        if result["OK"]:
            voName = result["Value"]

    records = []
    vmClient = VMClient()

    result = vmClient.getCEInstances(siteList, ceList, voName)
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACExit(-1)

    for nodeID in result["Value"]:
        nodeDict = result["Value"][nodeID]
        record = [
            nodeDict["Site"],
            nodeDict["CEName"],
            nodeID,
            nodeDict["NodeName"],
            nodeDict["PublicIP"],
            nodeDict["State"],
        ]
        records.append(record)

    fields = ["Site", "Endpoint", "ID", "Name", "PublicIP", "State"]
    printTable(fields, records)
    DIRACExit(0)


if __name__ == "__main__":
    main()



""" Test the FTS3Utilities"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import unittest

import mock

from DIRAC.DataManagementSystem.Client.FTS3File import FTS3File
from DIRAC import S_OK, S_ERROR

from DIRAC.DataManagementSystem.private.FTS3Utilities import groupFilesByTarget, selectUniqueSource, FTS3ServerPolicy
from DIRAC.DataManagementSystem.private.FTS3Plugins.DefaultFTS3Plugin import DefaultFTS3Plugin


def mock__checkSourceReplicas(ftsFiles, preferDisk=False):
    succ = {}
    failed = {}

    for ftsFile in ftsFiles:
        if hasattr(ftsFile, "fakeAttr_possibleSources"):
            succ[ftsFile.lfn] = dict.fromkeys(getattr(ftsFile, "fakeAttr_possibleSources"))
        else:
            failed[ftsFile.lfn] = "No such file or directory"

    return S_OK({"Successful": succ, "Failed": failed})


class TestFileGrouping(unittest.TestCase):
    """Testing all the grouping functions of FTS3Utilities"""

    def setUp(self):
        self.f1 = FTS3File()
        self.f1.fakeAttr_possibleSources = ["Src1", "Src2"]
        self.f1.lfn = "f1"
        self.f1.targetSE = "target1"

        self.f2 = FTS3File()
        self.f2.fakeAttr_possibleSources = ["Src2", "Src3"]
        self.f2.lfn = "f2"
        self.f2.targetSE = "target2"

        self.f3 = FTS3File()
        self.f3.fakeAttr_possibleSources = ["Src4"]
        self.f3.lfn = "f3"
        self.f3.targetSE = "target1"

        # File does not exist :-)
        self.f4 = FTS3File()
        self.f4.lfn = "f4"
        self.f4.targetSE = "target3"

        self.allFiles = [self.f1, self.f2, self.f3, self.f4]

    def test_01_groupFilesByTarget(self):

        # empty input
        self.assertTrue(groupFilesByTarget([])["Value"] == {})

        res = groupFilesByTarget(self.allFiles)

        self.assertTrue(res["OK"])

        groups = res["Value"]

        self.assertTrue(self.f1 in groups["target1"])
        self.assertTrue(self.f2 in groups["target2"])
        self.assertTrue(self.f3 in groups["target1"])
        self.assertTrue(self.f4 in groups["target3"])

    @mock.patch(
        "DIRAC.DataManagementSystem.private.FTS3Utilities._checkSourceReplicas", side_effect=mock__checkSourceReplicas
    )
    def test_04_selectUniqueSource(self, _mk_checkSourceReplicas):
        """Suppose they all go to the same target"""

        fts3Plugin = DefaultFTS3Plugin()
        res = selectUniqueSource(self.allFiles, fts3Plugin)

        self.assertTrue(res["OK"])

        uniqueSources, _failedFiles = res["Value"]

        # There should be only f1,f2 and f3
        allReturnedFiles = []
        existingFiles = [self.f1, self.f2, self.f3]
        for _srcSe, ftsFiles in uniqueSources.items():
            allReturnedFiles.extend(ftsFiles)

        # No files should be duplicated and all files should be there, except the non existing one
        self.assertEqual(len(existingFiles), len(allReturnedFiles))
        self.assertEqual(set(existingFiles), set(allReturnedFiles))

        filesInSrc1 = uniqueSources.get("Src1", [])
        filesInSrc2 = uniqueSources.get("Src2", [])
        filesInSrc3 = uniqueSources.get("Src3", [])

        filesInSrc4 = uniqueSources.get("Src4", [])
        # f1
        self.assertTrue(self.f1 in filesInSrc1 + filesInSrc2)
        self.assertTrue(self.f2 in filesInSrc2 + filesInSrc3)
        self.assertTrue(self.f3 in filesInSrc4)


def mock__failoverServerPolicy(_attempt):
    return "server_0"


def mock__randomServerPolicy(_attempt):
    return "server_0"


def mock__sequenceServerPolicy(_attempt):
    return "server_0"


def mock__OKFTSServerStatus(ftsServer):
    return S_OK(ftsServer)


def mock__ErrorFTSServerStatus(ftsServer):
    return S_ERROR(ftsServer)


class TestFTS3ServerPolicy(unittest.TestCase):
    """Testing FTS3 ServerPolicy selection"""

    def setUp(self):
        self.fakeServerDict = {
            "server_0": "server0.cern.ch",
            "server_1": "server1.cern.ch",
            "server_2": "server2.cern.ch",
        }

    @mock.patch(
        "DIRAC.DataManagementSystem.private.FTS3Utilities.FTS3ServerPolicy._getFTSServerStatus",
        side_effect=mock__OKFTSServerStatus,
    )
    @mock.patch(
        "DIRAC.DataManagementSystem.private.FTS3Utilities.FTS3ServerPolicy._sequenceServerPolicy",
        side_effect=mock__sequenceServerPolicy,
    )
    @mock.patch(
        "DIRAC.DataManagementSystem.private.FTS3Utilities.FTS3ServerPolicy._randomServerPolicy",
        side_effect=mock__randomServerPolicy,
    )
    @mock.patch(
        "DIRAC.DataManagementSystem.private.FTS3Utilities.FTS3ServerPolicy._failoverServerPolicy",
        side_effect=mock__failoverServerPolicy,
    )
    def testCorrectServerPolicyIsUsed(self, mockFailoverFunc, mockRandomFunc, mockSequenceFunc, mockFTSServerStatus):
        "Test correct server policy method is called"

        obj = FTS3ServerPolicy(self.fakeServerDict, "Sequence")
        obj.chooseFTS3Server()
        self.assertTrue(mockSequenceFunc.called)

        obj = FTS3ServerPolicy(self.fakeServerDict, "Random")
        obj.chooseFTS3Server()
        self.assertTrue(mockRandomFunc.called)

        obj = FTS3ServerPolicy(self.fakeServerDict, "Failover")
        obj.chooseFTS3Server()
        self.assertTrue(mockFailoverFunc.called)

        # random policy should be selected for an invalid policy
        obj = FTS3ServerPolicy(self.fakeServerDict, "InvalidPolicy")
        obj.chooseFTS3Server()
        self.assertTrue(mockRandomFunc.called)

    @mock.patch(
        "DIRAC.DataManagementSystem.private.FTS3Utilities.FTS3ServerPolicy._getFTSServerStatus",
        side_effect=mock__ErrorFTSServerStatus,
    )
    def testFailoverServerPolicy(self, mockFTSServerStatus):
        """Test if the failover server policy returns server at a given position"""

        obj = FTS3ServerPolicy(self.fakeServerDict, "Failover")
        for i in range(len(self.fakeServerDict)):
            self.assertEqual("server_%d" % i, obj._failoverServerPolicy(i))

    @mock.patch(
        "DIRAC.DataManagementSystem.private.FTS3Utilities.FTS3ServerPolicy._getFTSServerStatus",
        side_effect=mock__ErrorFTSServerStatus,
    )
    def testSequenceServerPolicy(self, mockFTSServerStatus):
        """Test if the sequence server policy selects the servers Sequentially"""

        obj = FTS3ServerPolicy(self.fakeServerDict, "Sequence")

        for i in range(len(self.fakeServerDict)):
            self.assertEqual("server_%d" % i, obj._sequenceServerPolicy(i))

        self.assertEqual("server_0", obj._sequenceServerPolicy(i))

    @mock.patch(
        "DIRAC.DataManagementSystem.private.FTS3Utilities.FTS3ServerPolicy._getFTSServerStatus",
        side_effect=mock__ErrorFTSServerStatus,
    )
    def testRandomServerPolicy(self, mockFTSServerStatus):
        """Test if the random server policy does not selects the same server multiple times"""

        obj = FTS3ServerPolicy(self.fakeServerDict, "Random")
        serverSet = set()

        for i in range(len(self.fakeServerDict)):
            serverSet.add(obj._randomServerPolicy(i))

        self.assertEqual(len(serverSet), len(self.fakeServerDict))


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestFileGrouping)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestFTS3ServerPolicy))
    unittest.TextTestRunner(verbosity=2).run(suite)



""" Definition for WMSHistory Monitoring type.
    Drop-in replacement for the Accounting/WMSHistory accounting type.

    Filled by the agent "WorkloadManagement/StatesAccountingAgent"
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.MonitoringSystem.Client.Types.BaseType import BaseType

__RCSID__ = "$Id$"


class WMSHistory(BaseType):

    """
    .. class:: WMSMonitorType
    """

    def __init__(self):
        """c'tor

        :param self: self reference
        """

        super(WMSHistory, self).__init__()

        self.keyFields = [
            "Status",
            "Site",
            "User",
            "UserGroup",
            "JobGroup",
            "MinorStatus",
            "ApplicationStatus",
            "JobSplitType",
        ]

        self.monitoringFields = ["Jobs", "Reschedules"]

        self.index = "wmshistory_index"

        self.addMapping(
            {
                "Status": {"type": "keyword"},
                "Site": {"type": "keyword"},
                "JobSplitType": {"type": "keyword"},
                "ApplicationStatus": {"type": "keyword"},
                "MinorStatus": {"type": "keyword"},
                "User": {"type": "keyword"},
                "JobGroup": {"type": "keyword"},
                "UserGroup": {"type": "keyword"},
            }
        )
        # {'timestamp': {'type': 'date'}} will be added for all monitoring types

        self.dataToKeep = 86400 * 30

        self.checkType()



"""
This module contains constants and lists for the possible job states.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.StateMachine import State, StateMachine


#:
SUBMITTING = "Submitting"
#:
RECEIVED = "Received"
#:
CHECKING = "Checking"
#:
STAGING = "Staging"
#:
WAITING = "Waiting"
#:
MATCHED = "Matched"
#: The Rescheduled status is effectively never stored in the DB.
#: It could be considered a "virtual" status, and might even be dropped.
RESCHEDULED = "Rescheduled"
#:
RUNNING = "Running"
#:
STALLED = "Stalled"
#:
COMPLETING = "Completing"
#:
DONE = "Done"
#:
COMPLETED = "Completed"
#:
FAILED = "Failed"
#:
DELETED = "Deleted"
#:
KILLED = "Killed"

#: Possible job states
JOB_STATES = [
    SUBMITTING,
    RECEIVED,
    CHECKING,
    STAGING,
    WAITING,
    MATCHED,
    RESCHEDULED,
    RUNNING,
    STALLED,
    COMPLETING,
    DONE,
    COMPLETED,
    FAILED,
    DELETED,
    KILLED,
]

# Job States when the payload work has finished
JOB_FINAL_STATES = [DONE, COMPLETED, FAILED]

# WMS internal job States indicating the job object won't be updated
JOB_REALLY_FINAL_STATES = [DELETED]


class JobsStateMachine(StateMachine):
    """Jobs state machine"""

    def __init__(self, state):
        """c'tor
        Defines the state machine transactions
        """
        super(JobsStateMachine, self).__init__(state)

        # States transitions
        self.states = {
            DELETED: State(14),  # final state
            KILLED: State(13, [DELETED], defState=KILLED),
            FAILED: State(12, [RESCHEDULED, DELETED], defState=FAILED),
            DONE: State(11, [DELETED], defState=DONE),
            COMPLETED: State(10, [DONE, FAILED], defState=COMPLETED),
            COMPLETING: State(9, [DONE, FAILED, COMPLETED], defState=COMPLETING),
            STALLED: State(8, [RUNNING, FAILED, KILLED], defState=STALLED),
            RUNNING: State(7, [STALLED, DONE, FAILED, COMPLETING, KILLED, RECEIVED], defState=RUNNING),
            RESCHEDULED: State(6, [WAITING, RECEIVED, DELETED], defState=RESCHEDULED),
            MATCHED: State(5, [RUNNING, FAILED, KILLED], defState=MATCHED),
            WAITING: State(4, [MATCHED, RESCHEDULED, DELETED], defState=WAITING),
            STAGING: State(3, [WAITING, FAILED, KILLED], defState=STAGING),
            CHECKING: State(2, [STAGING, WAITING, RESCHEDULED, FAILED, DELETED], defState=CHECKING),
            RECEIVED: State(1, [CHECKING, WAITING, FAILED, DELETED], defState=RECEIVED),
            SUBMITTING: State(0, [RECEIVED, CHECKING, DELETED], defState=SUBMITTING),  # initial state
        }



########################################################################
# File :   IdProviderFactory.py
# Author : A.T.
########################################################################

"""  The Identity Provider Factory instantiates IdProvider objects
     according to their configuration
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getInfoAboutProviders

__RCSID__ = "$Id$"


class IdProviderFactory(object):

    #############################################################################
    def __init__(self):
        """Standard constructor"""
        self.log = gLogger.getSubLogger("IdProviderFactory")

    #############################################################################
    def getIdProvider(self, idProvider):
        """This method returns a IdProvider instance corresponding to the supplied name.

        :param str idProvider: the name of the Identity Provider

        :return: S_OK(IdProvider)/S_ERROR()
        """
        result = getInfoAboutProviders(of="Id", providerName=idProvider, option="all", section="all")
        if not result["OK"]:
            return result
        pDict = result["Value"]
        pDict["ProviderName"] = idProvider
        pType = pDict["ProviderType"]

        self.log.verbose("Creating IdProvider", "of %s type with the name %s" % (pType, idProvider))
        subClassName = "%sIdProvider" % (pType)

        result = ObjectLoader().loadObject("Resources.IdProvider.%s" % subClassName)
        if not result["OK"]:
            self.log.error("Failed to load object", "%s: %s" % (subClassName, result["Message"]))
            return result

        pClass = result["Value"]
        try:
            provider = pClass()
            provider.setParameters(pDict)
        except Exception as x:
            msg = "IdProviderFactory could not instantiate %s object: %s" % (subClassName, str(x))
            self.log.exception()
            self.log.error(msg)
            return S_ERROR(msg)

        return S_OK(provider)



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-get-job-pilot-output
# Author :  Stuart Paterson
########################################################################
"""
Retrieve the output of the pilot that executed a given job

Example:
  $ dirac-admin-get-job-pilot-output 34
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC ID of the Job"])
    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import exit as DIRACExit
    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    errorList = []

    for job in args:

        try:
            job = int(job)
        except Exception as x:
            errorList.append(("Expected integer for JobID", job))
            exitCode = 2
            continue

        result = diracAdmin.getJobPilotOutput(job)
        if not result["OK"]:
            errorList.append((job, result["Message"]))
            exitCode = 2

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRACExit(exitCode)


if __name__ == "__main__":
    main()
